"""Common scheduler code for Adaptive Scheduler."""

from __future__ import annotations

import os
import os.path
import subprocess
import time

from rich.console import Console

console = Console()


def run_submit(cmd: str, name: str | None = None) -> None:
    """Run a submit command."""
    env = os.environ.copy()
    if name is not None:
        env["NAME"] = name
    for _ in range(10):
        proc = subprocess.run(cmd.split(), env=env, capture_output=True, check=False)
        if proc.returncode == 0:
            return
        stderr = proc.stderr.decode()
        if stderr:
            console.log(f"Error: {stderr}")
        time.sleep(0.5)



"""Test the widgets module."""

from __future__ import annotations

import asyncio
import os
import tempfile
from datetime import timedelta
from pathlib import Path
from random import randint
from typing import TYPE_CHECKING
from unittest.mock import patch

import pandas as pd

from adaptive_scheduler.widgets import (
    _bytes_to_human_readable,
    _failed_job_logs,
    _files_that_contain,
    _get_fnames,
    _interp_red_green,
    _sort_fnames,
    _timedelta_to_human_readable,
    _total_size,
    info,
    log_explorer,
    queue_widget,
    results_widget,
)

if TYPE_CHECKING:
    from typing import Any

    import pytest


def create_temp_files(num_files: int) -> list[str]:
    """Create a list of temporary files."""
    temp_files = []
    for _ in range(num_files):
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
            f.write("A" * randint(1, 100))  # noqa: S311
            temp_files.append(f.name)
    return temp_files


def _sum_size(files: list[str]) -> int:
    return sum(os.path.getsize(f) for f in files)  # noqa: PTH202


def test_total_size_single_list() -> None:
    """Test the _total_size function with a single list."""
    temp_files = create_temp_files(5)
    total_size = _sum_size(temp_files)
    assert _total_size(temp_files) == total_size


def test_total_size_nested_list() -> None:
    """Test the _total_size function with nested lists."""
    temp_files = create_temp_files(5)
    nested_files = [temp_files[:2], [temp_files[2]], temp_files[3:]]
    total_size = _sum_size(temp_files)
    assert _total_size(nested_files) == total_size


def test_total_size_mixed_types() -> None:
    """Test the _total_size function with mixed types in the list."""
    temp_files = create_temp_files(5)
    temp_paths = [Path(f) for f in temp_files[1:4]]
    mixed_files = [temp_files[0], temp_paths, temp_files[4]]
    total_size = _sum_size(temp_files)
    assert _total_size(mixed_files) == total_size  # type: ignore[arg-type]


def test_total_size_non_existing() -> None:
    """Test the _total_size function with non-existing files."""
    assert _total_size(["non_existing1", Path("Non_existing2")]) == 0  # type: ignore[arg-type]


def test_bytes_to_human_readable() -> None:
    """Test the _bytes_to_human_readable function."""
    size_in_bytes = 1234567890
    assert _bytes_to_human_readable(size_in_bytes) == "1.15 GB"


class MockDatabaseManager:
    """Mock DatabaseManager for testing."""

    def __init__(self, tmp_folder: Path) -> None:
        """Initialize the MockDatabaseManager."""
        self.failed: list[dict[str, Any]] = []
        self._total_learner_size = None
        self.fnames: list[str | Path] = []

        self.tmp_folder = tmp_folder

    def as_dicts(self) -> list[dict[str, Any]]:
        """Return a list of dictionaries representing the database entries."""
        return [
            {
                "fname": f"data/{i}.pickle",
                "job_id": f"100{i}",
                "is_done": False,
                "log_fname": f"{self.tmp_folder}/adaptive-{i}-100{i}.log",
                "job_name": f"adaptive-{i}",
                "output_logs": [f"{self.tmp_folder}/adaptive-{i}-100{i}.out"],
                "start_time": "2023-04-18T23:47:45.167896+00:00",
            }
            for i in range(3)
        ]

    def update(self, queue: dict) -> None:
        """Update the database with the given queue."""


class MockRunManager:
    """Mock RunManager for testing."""

    def __init__(self, tmp_folder: Path) -> None:
        """Initialize the MockRunManager."""
        self.job_name = "adaptive"
        self.scheduler = self
        self.log_folder = tmp_folder
        self.ext = ".log"
        self.move_old_logs_to = None
        self.max_log_lines = 500
        self._folder = tmp_folder
        self.database_manager = MockDatabaseManager(tmp_folder)
        self.job_names = ["adaptive-{i}" for i in range(3)]

    def status(self) -> str:
        """Return the status of the RunManager."""
        return "running"

    def elapsed_time(self) -> float:
        """Return the elapsed time of the RunManager."""
        return 123.0

    def job_starting_times(self) -> list[float]:
        """Return the starting times of the jobs."""
        return [float(i) for i in range(30)]

    def queue(
        self,
        me_only: bool = True,  # noqa: FBT001, ARG002, FBT002
    ) -> dict[str, dict[str, str]]:
        """Return a dictionary representing the current queue."""
        return {
            f"{i}": {
                "state": "RUNNING",
                "NumNodes": "1",
                "NumTasks": "1",
                "ReasonList": "barb8-nc24-high-1",
                "SubmitTime": "2023-04-16T22:09:53",
                "StartTime": "2023-04-16T22:09:54",
                "UserName": "basnijholt",
                "Partition": "nc24-high",
                "job_name": f"adaptive-{i}",
            }
            for i in range(3)
        }


def test_files_that_contain2(tmp_path: Path) -> None:
    """Test the _files_that_contain function."""
    d = tmp_path / "logs"
    d.mkdir()
    p1 = d / "test1.log"
    p1.write_text("Hello, world!")
    p2 = d / "test2.log"
    p2.write_text("Hello, everyone!")
    fnames = [p1, p2]
    result = _files_that_contain(fnames, "world")
    assert result == [p1]


def test_get_fnames(tmp_path: Path) -> None:
    """Test the _get_fnames function."""
    d = tmp_path / "logs"
    d.mkdir()
    p1 = d / "test_job-1.log"
    p1.write_text("Log 1")
    p2 = d / "test_job-2.log"
    p2.write_text("Log 2")

    run_manager = MockRunManager(tmp_path)

    fnames = _get_fnames(run_manager, only_running=False)  # type: ignore[arg-type]
    assert len(fnames) == 2

    fnames = _get_fnames(run_manager, only_running=True)  # type: ignore[arg-type]
    assert len(fnames) == 6


def test_failed_job_logs(tmp_path: Path) -> None:
    """Test the _failed_job_logs function."""
    d = tmp_path / "logs"
    d.mkdir()
    p1 = d / "test_job-1.log"
    p1.write_text("Log 1")
    p2 = d / "test_job-2.log"
    p2.write_text("Log 2")

    run_manager = MockRunManager(tmp_path)

    fnames = [p1, p2]
    failed_fnames = _failed_job_logs(fnames, run_manager, only_running=False)  # type: ignore[arg-type]
    assert failed_fnames == []


def test_timedelta_to_human_readable_int() -> None:
    """Test the _timedelta_to_human_readable function with an integer."""
    seconds = 3666
    assert _timedelta_to_human_readable(seconds) == "1 h, 1 m, 6 s"
    assert (
        _timedelta_to_human_readable(seconds, short_format=False) == "1 hour, 1 minute, 6 seconds"
    )


def test_log_explorer(tmp_path: Path) -> None:
    """Test the log_explorer function."""
    run_manager = MockRunManager(tmp_path)

    async def mock_get_running_loop():  # noqa: ANN202
        return asyncio.get_event_loop()

    with patch("asyncio.get_running_loop", side_effect=mock_get_running_loop):
        widget = log_explorer(run_manager)  # type: ignore[arg-type]

    assert widget is not None


def test_queue_widget(tmp_path: Path) -> None:
    """Test the queue_widget function."""
    run_manager = MockRunManager(tmp_path)
    widget = queue_widget(run_manager)  # type: ignore[arg-type]
    assert widget is not None


def test_info(capfd: pytest.CaptureFixture, tmp_path: Path) -> None:
    """Test the info function."""
    run_manager = MockRunManager(tmp_path)
    info(run_manager)  # type: ignore[arg-type]
    captured = capfd.readouterr()
    assert "status" in captured.out


def test_timedelta_to_human_readable_short_format() -> None:
    """Test the _timedelta_to_human_readable function with short_format."""
    time_input = timedelta(days=2, hours=3, minutes=4, seconds=5)
    assert _timedelta_to_human_readable(time_input, short_format=True) == "2 d, 3 h, 4 m, 5 s"


def test_files_that_contain(tmp_path: Path) -> None:
    """Test the _files_that_contain function."""
    file1 = tmp_path / "file1.txt"
    file1.write_text("This file contains the keyword: example")

    file2 = tmp_path / "file2.txt"
    file2.write_text("This file does not contain the keyword.")

    fnames = [file1, file2]
    filtered_fnames = _files_that_contain(fnames, "example")
    assert len(filtered_fnames) == 1
    assert filtered_fnames[0] == file1


def test_get_fnames_only_running(tmp_path: Path) -> None:
    """Test the _get_fnames function with only_running=True."""
    run_manager = MockRunManager(tmp_path)
    fnames = _get_fnames(run_manager, only_running=True)  # type: ignore[arg-type]
    assert len(fnames) == 6


def test_get_fnames_only_running_false(tmp_path: Path) -> None:
    """Test the _get_fnames function with only_running=False."""
    run_manager = MockRunManager(tmp_path)
    log_fname = run_manager.database_manager.as_dicts()[0]["log_fname"]
    with open(log_fname, "w") as f:  # noqa: PTH123
        f.write("This is a test log file.")
    fnames = _get_fnames(run_manager, only_running=False)  # type: ignore[arg-type]
    assert len(fnames) == 3
    # TODO: it is picking up the database file and scheduler file
    # we should probably filter those out?


def test_sort_fnames(tmp_path: Path) -> None:
    """Test the _sort_fnames function."""
    run_manager = MockRunManager(tmp_path)
    fnames = [Path(f"logs/{run_manager.job_name}-{i}-{i * 100}.log") for i in range(3)]

    sorted_fnames = _sort_fnames("Alphabetical", run_manager, fnames)  # type: ignore[arg-type]
    assert sorted_fnames == fnames  # In this case, they should be the same


def test_results_widget(tmp_path: Path) -> None:
    """Test the results_widget function."""
    # Create some sample data files in DataFrame pickle format
    sample_data = [
        pd.DataFrame({"col1": range(10), "col2": range(10)}),
        pd.DataFrame({"col1": range(10, 20), "col2": range(10, 20)}),
    ]

    data_files = []
    for idx, data in enumerate(sample_data):
        file_path = tmp_path / f"sample_data_{idx}.pickle"
        data.to_pickle(file_path)
        data_files.append(file_path)

    # Create the widget
    widget = results_widget(data_files, "pickle")

    # Test the widget's configuration
    assert widget is not None


def test_interp_red_green() -> None:
    """Test _interp_red_green."""
    assert _interp_red_green(100, pct_red=10, pct_green=90) == (0, 255, 0)
    assert _interp_red_green(0, pct_red=10, pct_green=90) == (255, 0, 0)
    assert _interp_red_green(50, pct_red=10, pct_green=90) == (127, 127, 0)
    assert _interp_red_green(50, pct_red=90, pct_green=10) == (127, 127, 0)
    assert _interp_red_green(100, pct_green=10, pct_red=90) == (255, 0, 0)
    assert _interp_red_green(0, pct_green=10, pct_red=90) == (0, 255, 0)



"""Tests for the slurm_run function."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

from adaptive_scheduler._server_support.run_manager import RunManager
from adaptive_scheduler._server_support.slurm_run import slurm_run
from adaptive_scheduler.scheduler import SLURM

if TYPE_CHECKING:
    from pathlib import Path

    import adaptive

    from adaptive_scheduler.utils import _DATAFRAME_FORMATS


@pytest.fixture
def extra_run_manager_kwargs() -> dict[str, Any]:
    """Fixture for creating extra run manager keyword arguments."""
    return {"kill_on_error": "GPU on fire", "loky_start_method": "fork"}


@pytest.fixture
def extra_scheduler_kwargs() -> dict[str, Any]:
    """Fixture for creating extra scheduler keyword arguments."""
    return {"mpiexec_executable": "mpiexec"}


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_slurm_run_with_default_arguments(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test slurm_run function with default arguments."""
    rm = slurm_run(learners, fnames)
    assert isinstance(rm, RunManager)
    assert isinstance(rm.scheduler, SLURM)
    assert rm.save_interval == 300
    assert rm.log_interval == 300
    assert rm.learners == learners
    assert rm.fnames == fnames
    assert rm.scheduler.exclusive is False


def goal_example(learner: adaptive.Learner1D) -> bool:
    """Example goal function for testing."""
    return len(learner.data) >= 10


@pytest.mark.parametrize(
    ("partition", "nodes", "cores_per_node"),
    [("hb120v2-low", 2, 120), ("hb60-high", 3, 30)],
)
@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_slurm_run_with_custom_partition_nodes_and_cores(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
    partition: str,
    nodes: int,
    cores_per_node: int,
) -> None:
    """Test slurm_run function with custom partition, nodes, and cores_per_node."""
    rm = slurm_run(
        learners,
        fnames,
        partition=partition,
        nodes=nodes,
        cores_per_node=cores_per_node,
        executor_type="ipyparallel",
    )
    assert isinstance(rm, RunManager)
    assert isinstance(rm.scheduler, SLURM)
    assert rm.scheduler.partition == partition
    assert rm.scheduler.nodes == nodes
    assert rm.scheduler.cores_per_node == cores_per_node


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_slurm_run_with_custom_goal(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test slurm_run function with custom goal."""
    rm = slurm_run(learners, fnames, goal=goal_example)
    assert isinstance(rm, RunManager)
    assert rm.goal == goal_example


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_slurm_run_with_custom_folder_and_name(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
    tmp_path: Path,
) -> None:
    """Test slurm_run function with custom folder and name."""
    folder = tmp_path / "test_folder"
    name = "custom_name"
    rm = slurm_run(learners, fnames, folder=folder, name=name)
    assert isinstance(rm, RunManager)
    assert rm.scheduler.log_folder == folder / "logs"
    assert rm.job_name == name


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_slurm_run_with_custom_num_threads(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test slurm_run function with custom num_threads."""
    num_threads = 4
    rm = slurm_run(learners, fnames, num_threads=num_threads)
    assert isinstance(rm, RunManager)
    assert rm.scheduler.num_threads == num_threads


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_slurm_run_with_extra_run_manager_kwargs(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
    extra_run_manager_kwargs: dict[str, Any],
) -> None:
    """Test slurm_run function with extra_run_manager_kwargs."""
    rm = slurm_run(learners, fnames, extra_run_manager_kwargs=extra_run_manager_kwargs)
    assert isinstance(rm, RunManager)
    for key, value in extra_run_manager_kwargs.items():
        assert getattr(rm, key) == value


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
@pytest.mark.parametrize("dataframe_format", ["csv", "json", "pickle"])
def test_slurm_run_with_custom_dataframe_format(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
    dataframe_format: _DATAFRAME_FORMATS,
) -> None:
    """Test slurm_run function with custom dataframe_format."""
    rm = slurm_run(learners, fnames, dataframe_format=dataframe_format)
    assert isinstance(rm, RunManager)
    assert rm.dataframe_format == dataframe_format


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_slurm_run_with_custom_max_fails_and_jobs(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test slurm_run function with custom max_fails_per_job and max_simultaneous_jobs."""
    max_fails_per_job = 100
    max_simultaneous_jobs = 200
    rm = slurm_run(
        learners,
        fnames,
        max_fails_per_job=max_fails_per_job,
        max_simultaneous_jobs=max_simultaneous_jobs,
    )
    assert isinstance(rm, RunManager)
    assert rm.max_fails_per_job == max_fails_per_job
    assert rm.max_simultaneous_jobs == max_simultaneous_jobs


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_slurm_run_with_extra_scheduler_kwargs(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
    extra_scheduler_kwargs: dict[str, Any],
) -> None:
    """Test slurm_run function with extra_scheduler_kwargs."""
    rm = slurm_run(learners, fnames, extra_scheduler_kwargs=extra_scheduler_kwargs)
    assert isinstance(rm, RunManager)
    for key, value in extra_scheduler_kwargs.items():
        assert getattr(rm.scheduler, key) == value


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
@pytest.mark.parametrize("executor_type", ["ipyparallel", "dask-mpi", "mpi4py"])
def test_slurm_run_with_custom_executor_type(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
    executor_type: str,
) -> None:
    """Test slurm_run function with custom executor_type."""
    rm = slurm_run(
        learners,
        fnames,
        executor_type=executor_type,  # type: ignore[arg-type]
    )
    assert isinstance(rm, RunManager)
    assert rm.scheduler.executor_type == executor_type


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_slurm_run_with_invalid_nodes_and_executor_type(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test slurm_run function with invalid nodes and executor_type."""
    with pytest.raises(
        ValueError,
        match="process-pool can maximally use a single node",
    ):
        slurm_run(learners, fnames, nodes=2, executor_type="process-pool")


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_slurm_run_with_multiple_partitions(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test slurm_run function with multiple partitions."""
    rm = slurm_run(learners, fnames, partition=("hb120v2-low", "hb60-high"))
    assert isinstance(rm, RunManager)
    s = rm.scheduler
    assert isinstance(s, SLURM)
    assert s.cores == (120, 60)
    assert s.partition == ("hb120v2-low", "hb60-high")
    assert s.nodes == (1, 1)
    assert s.cores_per_node == (120, 60)



"""Test module for JobManager."""

from __future__ import annotations

import asyncio
import logging
from typing import Any
from unittest.mock import MagicMock

import pytest

from adaptive_scheduler.server_support import JobManager, MaxRestartsReachedError


@pytest.mark.asyncio
async def test_job_manager_init(job_manager: JobManager) -> None:
    """Test the initialization of JobManager."""
    job_manager.database_manager.start()
    job_manager.start()
    assert job_manager.task is not None


@pytest.mark.asyncio
async def test_job_manager_queued(job_manager: JobManager) -> None:
    """Test the _queued method of JobManager."""
    job_manager.scheduler.start_job("job1")
    job_manager.scheduler.start_job("job2")
    job_manager.scheduler.update_queue("job1", "running")  # type: ignore[attr-defined]
    job_manager.scheduler.update_queue("job2", "running")  # type: ignore[attr-defined]
    assert job_manager._queued(job_manager.scheduler.queue()) == {"job1", "job2"}


@pytest.mark.asyncio
async def test_job_manager_manage_max_restarts_reached(job_manager: JobManager) -> None:
    """Test the JobManager when the maximum restarts are reached."""
    job_manager.n_started = 105
    # Should fail after n_started > n_learners * max_fails_per_job
    # Which is `105 > 2 * 50 = True`
    job_manager.scheduler._queue_info = {}  # type: ignore[attr-defined]
    job_manager.database_manager.start()
    job_manager.start()
    assert job_manager.task is not None
    await asyncio.sleep(0.1)
    with pytest.raises(
        MaxRestartsReachedError,
        match="Too many jobs failed, your Python code probably has a bug",
    ):
        job_manager.task.result()


@pytest.mark.asyncio
async def test_job_manager_manage_start_jobs(job_manager: JobManager) -> None:
    """Test the JobManager when managing the start of jobs."""
    job_manager.database_manager.n_done = MagicMock(return_value=0)  # type: ignore[method-assign]
    job_manager.scheduler._queue_info = {}  # type: ignore[attr-defined]
    job_manager.max_simultaneous_jobs = 2
    job_manager.database_manager.start()
    job_manager.start()
    await asyncio.sleep(0.1)
    assert set(job_manager.scheduler._started_jobs) == {"job1", "job2"}  # type: ignore[attr-defined]


@pytest.mark.asyncio
async def test_job_manager_manage_start_max_simultaneous_jobs(
    job_manager: JobManager,
) -> None:
    """Test the JobManager when managing the maximum simultaneous jobs."""
    job_manager.database_manager.n_done = MagicMock(return_value=0)  # type: ignore[method-assign]
    job_manager.scheduler._queue_info = {}  # type: ignore[attr-defined]
    job_manager.interval = 0.1  # type: ignore[assignment]
    job_manager.max_simultaneous_jobs = 1
    job_manager.database_manager.start()
    job_manager.start()
    assert job_manager.task is not None
    await asyncio.sleep(0.15)
    assert len(job_manager.scheduler._started_jobs) == 1  # type: ignore[attr-defined]


@pytest.mark.asyncio
async def test_job_manager_manage_cancelled_error(
    job_manager: JobManager,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test the JobManager when a CancelledError occurs."""
    caplog.set_level(logging.INFO)
    job_manager.database_manager.start()
    job_manager.start()
    assert job_manager.task is not None

    timeout = 0.1
    await asyncio.sleep(timeout)
    assert job_manager.database_manager.task is not None
    assert not job_manager.database_manager.task.done()
    assert not job_manager.task.done()
    job_manager.task.cancel()
    await asyncio.sleep(timeout)
    with pytest.raises(asyncio.CancelledError):
        await job_manager.task

    assert "task was cancelled because of a CancelledError" in caplog.text


@pytest.mark.asyncio
async def test_job_manager_manage_n_done_equal_job_names(
    job_manager: JobManager,
) -> None:
    """Test the JobManager when n_done equals the number of job names."""
    job_manager.database_manager.n_done = MagicMock(  # type: ignore[method-assign]
        return_value=len(job_manager.job_names),
    )
    job_manager.database_manager.start()
    job_manager.start()
    assert job_manager.task is not None
    await asyncio.sleep(0.1)
    assert job_manager.task.done()
    assert job_manager.task.result() is None


@pytest.mark.asyncio
async def test_job_manager_manage_generic_exception(
    job_manager: JobManager,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test the JobManager when a generic exception occurs."""

    def raise_exception(*_args: Any, **_kwargs: Any) -> None:
        msg = "Test exception"
        raise ValueError(msg)

    caplog.set_level(logging.ERROR)
    job_manager.scheduler.start_job = MagicMock(side_effect=raise_exception)  # type: ignore[method-assign]
    job_manager.database_manager.start()
    job_manager.start()
    assert job_manager.task is not None
    await asyncio.sleep(0.15)
    assert "got exception when starting a job" in caplog.text
    assert "Test exception" in caplog.text



from __future__ import annotations

import asyncio
import itertools
import logging
import shutil
import socket
from concurrent.futures import ThreadPoolExecutor
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING

import structlog
import zmq
import zmq.asyncio
import zmq.ssh
from rich.console import Console

from adaptive_scheduler.utils import (
    _progress,
    _remove_or_move_files,
)

if TYPE_CHECKING:
    from adaptive_scheduler.scheduler import BaseScheduler

console = Console()

logger = logging.getLogger("adaptive_scheduler.server")
logger.setLevel(logging.INFO)
log = structlog.wrap_logger(logger)


class MaxRestartsReachedError(Exception):
    """Max restarts reached.

    Jobs can fail instantly because of an error in
    your Python code which results jobs being started indefinitely.
    """


def get_allowed_url() -> str:
    """Get an allowed url for the database manager.

    Returns
    -------
    url
        An url that can be used for the database manager, with the format
        ``tcp://ip_of_this_machine:allowed_port.``.

    """
    ip = socket.gethostbyname(socket.gethostname())
    port = zmq.ssh.tunnel.select_random_ports(1)[0]
    return f"tcp://{ip}:{port}"


def _get_matching_files(f: Path, variable: str) -> list[Path]:
    modified_name = f.name.replace(variable, "*")
    return list(f.parent.glob(modified_name))


def _get_all_files(job_names: list[str], scheduler: BaseScheduler) -> set[Path]:
    log_fnames = [scheduler.log_fname(name) for name in job_names]
    _output_fnames = [scheduler.output_fnames(name) for name in job_names]
    output_fnames: list[Path] = list(itertools.chain(*_output_fnames))
    batch_fnames = [scheduler.batch_fname(name) for name in job_names]
    fnames = log_fnames + output_fnames + batch_fnames
    all_files = []
    for f in fnames:
        matching_files = _get_matching_files(f, scheduler._JOB_ID_VARIABLE)
        all_files.extend(matching_files)
    # For schedulers that use a single batch file
    name_prefix = job_names[0].rsplit("-", 1)[0]
    batch_file = scheduler.batch_fname(name_prefix)
    if batch_file.exists():
        all_files.append(batch_file)
    return set(all_files)


def cleanup_scheduler_files(
    job_names: list[str],
    scheduler: BaseScheduler,
    *,
    with_progress_bar: bool = True,
    move_to: str | Path | None = None,
) -> None:
    """Cleanup the scheduler log-files files.

    Parameters
    ----------
    job_names
        List of job names.
    scheduler
        A scheduler instance from `adaptive_scheduler.scheduler`.
    with_progress_bar
        Display a progress bar using `tqdm`.
    move_to
        Move the file to a different directory.
        If None the file is removed.
    log_file_folder
        The folder in which to delete the log-files.

    """
    to_rm = _get_all_files(job_names, scheduler)

    _remove_or_move_files(
        to_rm,
        with_progress_bar=with_progress_bar,
        move_to=move_to,
        desc="Removing logs and batch files",
    )


IPYTHON_PROFILE_PATTERN = "profile_adaptive_scheduler_"


def _ipython_profiles() -> list[Path]:
    return list(Path("~/.ipython/").expanduser().glob(f"{IPYTHON_PROFILE_PATTERN}*"))


def _delete_old_ipython_profiles(
    scheduler: BaseScheduler,
    *,
    with_progress_bar: bool = True,
) -> None:
    # We need the job_ids because only job_names wouldn't be
    # enough information. There might be other job_managers
    # running.

    running_job_ids = set(scheduler.queue().keys())
    to_delete = [
        folder
        for folder in _ipython_profiles()
        if str(folder).split(IPYTHON_PROFILE_PATTERN)[1] not in running_job_ids
    ]

    with ThreadPoolExecutor(256) as ex:
        desc = "Submitting deleting old IPython profiles tasks"
        pbar = _progress(to_delete, desc=desc)
        futs = [ex.submit(shutil.rmtree, folder) for folder in pbar]
        desc = "Finishing deleting old IPython profiles"
        for fut in _progress(futs, with_progress_bar, desc=desc):
            fut.result()


def periodically_clean_ipython_profiles(
    scheduler: BaseScheduler,
    interval: float = 600,
) -> asyncio.Task:
    """Periodically remove old IPython profiles.

    In the `RunManager.cleanup` method the profiles will be removed. However,
    one might want to remove them earlier.

    Parameters
    ----------
    scheduler
        A scheduler instance from `adaptive_scheduler.scheduler`.
    interval
        The interval at which to remove old profiles.

    Returns
    -------
    asyncio.Task

    """
    if isinstance(scheduler.executor_type, tuple):
        msg = (
            "This function is not implemented for multiple executors."
            " Please open an issue on GitHub if you need this feature."
        )
        raise NotImplementedError(msg)

    async def clean(interval: float) -> None:
        while True:
            with suppress(Exception):
                _delete_old_ipython_profiles(scheduler, with_progress_bar=False)
            await asyncio.sleep(interval)

    ioloop = asyncio.get_event_loop()
    coro = clean(interval)
    return ioloop.create_task(coro)


def _maybe_path(fname: str | Path | None) -> Path | None:  # pragma: no cover
    """Convert a string to a Path or return None."""
    if fname is None:
        return None
    if isinstance(fname, str):
        return Path(fname)
    return fname



"""Utility functions for adaptive_scheduler."""

from __future__ import annotations

import asyncio
import base64
import functools
import hashlib
import inspect
import math
import os
import pickle
import platform
import random
import shutil
import tempfile
import time
import uuid
import warnings
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager, suppress
from datetime import datetime, timedelta, timezone
from inspect import signature
from itertools import chain
from multiprocessing import Manager
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
)
from typing import (
    get_args as get_type_args,
)

import adaptive
import cloudpickle
import numpy as np
import pandas as pd
import toolz
from adaptive.notebook_integration import in_ipynb
from rich.console import Console
from rich.progress import (
    Progress,
    TimeElapsedColumn,
    get_console,
)
from tqdm import tqdm, tqdm_notebook

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from multiprocessing.managers import ListProxy

    from rich.progress import TaskID

console = Console()

MAX_LINE_LENGTH = 100
_NONE_RETURN_STR = "__ReturnsNone__"
FnamesTypes = list[str] | list[Path] | list[list[str]] | list[list[Path]]

LOKY_START_METHODS = Literal[
    "loky",
    "loky_int_main",
    "spawn",
    "fork",
    "forkserver",
]

EXECUTOR_TYPES = Literal[
    "mpi4py",
    "ipyparallel",
    "dask-mpi",
    "process-pool",
    "loky",
    "sequential",
]
GoalTypes = Callable[[adaptive.BaseLearner], bool] | int | float | datetime | timedelta | None


def shuffle_list(*lists: list, seed: int | None = 0) -> zip:
    """Shuffle multiple lists in the same order."""
    combined = list(zip(*lists, strict=True))
    random.Random(seed).shuffle(combined)  # noqa: S311
    return zip(*combined, strict=True)


def hash_anything(x: Any) -> str:
    """Hash anything."""
    try:
        return hashlib.md5(x).hexdigest()  # noqa: S324
    except TypeError:
        return hashlib.md5(pickle.dumps(x)).hexdigest()  # noqa: S324


def split(seq: Iterable, n_parts: int) -> Iterable[tuple]:
    """Split up a sequence into ``n_parts``.

    Parameters
    ----------
    seq
        A list or other iterable that has to be split up.
    n_parts
        The sequence will be split up in this many parts.

    """
    lst = list(seq)
    n = math.ceil(len(lst) / n_parts)
    return toolz.partition_all(n, lst)


def split_in_balancing_learners(
    learners: list[adaptive.BaseLearner],
    fnames: list[str] | list[Path],
    n_parts: int,
    strategy: str = "npoints",
) -> tuple[list[adaptive.BaseLearner], list[list[str]] | list[list[Path]]]:
    r"""Split a list of learners and fnames into `adaptive.BalancingLearner`\s.

    Parameters
    ----------
    learners
        List of learners.
    fnames
        List of filenames.
    n_parts
        Total number of `~adaptive.BalancingLearner`\s.
    strategy
        Learning strategy of the `~adaptive.BalancingLearner`.

    Returns
    -------
    new_learners, new_fnames

    """
    new_learners = []
    new_fnames = []
    for x in split(zip(learners, fnames, strict=True), n_parts):
        learners_part, fnames_part = zip(*x, strict=True)
        learner = adaptive.BalancingLearner(learners_part, strategy=strategy)
        new_learners.append(learner)
        new_fnames.append(list(fnames_part))
    return new_learners, new_fnames


def split_sequence_learner(
    big_learner: adaptive.SequenceLearner,
    n_learners: int,
    folder: str | Path = "",
) -> tuple[list[adaptive.SequenceLearner], list[str]]:
    r"""Split a sinlge `~adaptive.SequenceLearner` into many.

    Split into mutiple `adaptive.SequenceLearner`\s
    (with the data loaded) and fnames.

    See also `split_sequence_in_sequence_learners`.

    Parameters
    ----------
    big_learner
        A `~adaptive.SequenceLearner` instance
    n_learners
        Total number of `~adaptive.SequenceLearner`\s.
    folder
        Folder to prepend to fnames.

    Returns
    -------
    new_learners
        List of `~adaptive.SequenceLearner`\s.
    new_fnames
        List of str based on a hash of the sequence.

    """
    new_learners, new_fnames = split_sequence_in_sequence_learners(
        function=big_learner._original_function,
        sequence=big_learner.sequence,
        n_learners=n_learners,
        folder=folder,
    )
    # Load the new learners with data
    index_parts = split(range(len(big_learner.sequence)), n_learners)
    for small_learner, part in zip(new_learners, index_parts, strict=True):
        for i_small, i_big in enumerate(part):
            y = big_learner.data.get(i_big)
            if y is None:
                continue
            x = i_small, big_learner.sequence[i_big]
            small_learner.tell(x, y)

    return new_learners, new_fnames


def split_sequence_in_sequence_learners(
    function: Callable[[Any], Any],
    sequence: Sequence[Any],
    n_learners: int,
    folder: str | Path = "",
) -> tuple[list[adaptive.SequenceLearner], list[str]]:
    r"""Split a sequenceinto `adaptive.SequenceLearner`\s and fnames.

    Parameters
    ----------
    function
        Function for `adaptive.SequenceLearner`\s.
    sequence
        The sequence to split into ``n_learners``.
    n_learners
        Total number of `~adaptive.SequenceLearner`\s.
    folder
        Folder to prepend to fnames.

    Returns
    -------
    new_learners
        List of `~adaptive.SequenceLearner`\s.
    new_fnames
        List of str based on a hash of the sequence.

    """
    folder = Path(folder)
    new_learners = []
    new_fnames = []
    for sequence_part in split(sequence, n_learners):
        learner = adaptive.SequenceLearner(function, sequence_part)
        new_learners.append(learner)
        hsh = hash_anything((sequence_part[0], len(sequence_part)))
        fname = folder / f"{hsh}.pickle"
        new_fnames.append(str(fname))
    return new_learners, new_fnames


def combine_sequence_learners(
    learners: list[adaptive.SequenceLearner],
    big_learner: adaptive.SequenceLearner | None = None,
) -> adaptive.SequenceLearner:
    r"""Combine several `~adaptive.SequenceLearner`\s into a single one.

    Also copy over the data.

    Assumes that all ``learners`` take the same function.

    Parameters
    ----------
    learners
        List of `~adaptive.SequenceLearner`\s.
    big_learner
        A learner to load, if None, a new learner will be generated.

    Returns
    -------
    adaptive.SequenceLearner
        Big `~adaptive.SequenceLearner` with data from ``learners``.

    """
    if big_learner is None:
        big_sequence: list[Any] = list(
            chain.from_iterable(learner.sequence for learner in learners),
        )
        big_learner = adaptive.SequenceLearner(
            learners[0]._original_function,
            sequence=big_sequence,
        )

    cnt = 0
    for learner in learners:
        for i, key in enumerate(learner.sequence):
            if i in learner.data:
                x = cnt, key
                y = learner.data[i]
                big_learner.tell(x, y)
            cnt += 1
    return big_learner


def copy_from_sequence_learner(
    learner_from: adaptive.SequenceLearner,
    learner_to: adaptive.SequenceLearner,
) -> None:
    """Copy the data from a `~adaptive.SequenceLearner` into a different one.

    Parameters
    ----------
    learner_from
        Learner to take the data from.
    learner_to
        Learner to tell the data to.

    """
    mapping = {hash_anything(learner_from.sequence[i]): v for i, v in learner_from.data.items()}
    for i, key in enumerate(learner_to.sequence):
        hsh = hash_anything(key)
        if hsh in mapping:
            v = mapping[hsh]
            learner_to.tell((i, key), v)


def _get_npoints(learner: adaptive.BaseLearner) -> int | None:
    with suppress(AttributeError):
        return learner.npoints
    with suppress(AttributeError):
        # If the Learner is a BalancingLearner
        return sum(learner.npoints for learner in learner.learners)
    return None


def _progress(
    seq: Iterable[Any],
    with_progress_bar: bool = True,  # noqa: FBT001, FBT002
    desc: str = "",
) -> Iterable | tqdm:
    if not with_progress_bar:
        return seq
    if in_ipynb():
        return tqdm_notebook(list(seq), desc=desc)
    return tqdm(list(seq), desc=desc)


def combo_to_fname(
    combo: dict[str, Any],
    folder: str | Path | None = None,
    ext: str = ".pickle",
) -> Path:
    """Converts a dict into a human readable filename."""
    console.log(
        "Use `adaptive_scheduler.utils.combo2fname` instead of `combo_to_fname`.",
    )
    fname = "__".join(f"{k}_{v}" for k, v in combo.items()) + ext
    if folder is None:
        return Path(fname)
    return Path(folder) / fname


def combo2fname(
    combo: dict[str, Any],
    folder: str | Path | None = None,
    ext: str = ".pickle",
    sig_figs: int = 8,
) -> Path:
    """Converts a dict into a human readable filename.

    Improved version of `combo_to_fname`.
    """
    name_parts = [f"{k}_{maybe_round(v, sig_figs)}" for k, v in sorted(combo.items())]
    fname = Path("__".join(name_parts) + ext)
    if folder is None:
        return fname
    return folder / fname


def add_constant_to_fname(
    combo: dict[str, Any],
    constant: dict[str, Any],
    *,
    folder: str | Path | None = None,
    ext: str = ".pickle",
    sig_figs: int = 8,
) -> tuple[Path, Path]:
    """Construct old and new filename based on a combo.

    Assumes `combo2fname` has been used to construct the old filename.
    Adds `constant` dict to the `combo` and returns the new filename too.

    Returns a tuple of ``old_fname`` and ``new_fname``.
    """
    for k in constant:
        combo.pop(k, None)
    old_fname = combo2fname(combo, folder, ext, sig_figs)
    combo.update(constant)
    new_fname = combo2fname(combo, folder, ext, sig_figs)
    return old_fname, new_fname


def maybe_round(x: Any, sig_figs: int) -> Any:
    """Round to specified number of sigfigs if x is a float or complex."""
    rnd = functools.partial(round_sigfigs, sig_figs=sig_figs)

    def try_is_nan_inf(x: Any) -> bool:
        try:
            return np.isnan(x) or np.isinf(x)
        except Exception:  # noqa: BLE001
            return False

    if try_is_nan_inf(x):
        return x
    if isinstance(x, float):
        return rnd(x)
    if isinstance(x, complex):
        return complex(rnd(x.real), rnd(x.imag))
    return x


def round_sigfigs(num: float, sig_figs: int) -> float:
    """Round to specified number of sigfigs.

    From
    http://code.activestate.com/recipes/578114-round-number-to-specified-number-of-significant-di/
    """
    num = float(num)
    if num != 0:
        return round(num, -int(math.floor(math.log10(abs(num))) - (sig_figs - 1)))
    return 0.0  # Can't take the log of 0


def _remove_or_move_files(
    fnames: Sequence[str] | set[str] | Sequence[Path] | set[Path],
    *,
    with_progress_bar: bool = True,
    move_to: str | Path | None = None,
    desc: str | None = None,
) -> None:
    """Remove files by filename.

    Parameters
    ----------
    fnames
        List of filenames.
    with_progress_bar
        Display a progress bar using `tqdm`.
    move_to
        Move the file to a different directory.
        If None the file is removed.
    desc
        Description of the progressbar.

    """
    n_failed = 0
    for fname in _progress(fnames, with_progress_bar, desc or "Removing files"):
        fname = Path(fname)  # noqa: PLW2901
        try:
            if move_to is None:
                fname.unlink()
            else:
                move_to = Path(move_to)
                move_to.mkdir(parents=True, exist_ok=True)
                src = fname.resolve()
                dst = (move_to / src.name).resolve()
                shutil.move(src, dst)  # overwrites old files
        except Exception:  # noqa: BLE001
            n_failed += 1

    if n_failed:
        warnings.warn(
            f"Failed to remove (or move) {n_failed}/{len(fnames)} files.",
            stacklevel=2,
        )


def load_parallel(
    learners: list[adaptive.BaseLearner],
    fnames: list[str] | list[Path],
    *,
    with_progress_bar: bool = True,
    max_workers: int | None = None,
) -> None:
    r"""Load a sequence of learners in parallel.

    Parameters
    ----------
    learners
        The learners to be loaded.
    fnames
        A list of filenames corresponding to `learners`.
    with_progress_bar
        Display a progress bar using `tqdm`.
    max_workers
        The maximum number of parallel threads when loading the data.
        If ``None``, use the maximum number of threads that is possible.

    """

    def load(learner: adaptive.BaseLearner, fname: str) -> None:
        learner.load(fname)

    with ThreadPoolExecutor(max_workers) as ex:
        iterator = zip(learners, fnames, strict=True)
        pbar = _progress(iterator, with_progress_bar, "Submitting loading tasks")
        futs = [ex.submit(load, *args) for args in pbar]
        for fut in _progress(futs, with_progress_bar, "Finishing loading"):
            fut.result()


def save_parallel(
    learners: list[adaptive.BaseLearner],
    fnames: list[str] | list[Path],
    *,
    with_progress_bar: bool = True,
) -> None:
    r"""Save a sequence of learners in parallel.

    Parameters
    ----------
    learners
        The learners to be saved.
    fnames
        A list of filenames corresponding to `learners`.
    with_progress_bar
        Display a progress bar using `tqdm`.

    """

    def save(learner: adaptive.BaseLearner, fname: str) -> None:
        learner.save(fname)

    with ThreadPoolExecutor() as ex:
        iterator = zip(learners, fnames, strict=True)
        pbar = _progress(iterator, with_progress_bar, "Submitting saving tasks")
        futs = [ex.submit(save, *args) for args in pbar]
        for fut in _progress(futs, with_progress_bar, "Finishing saving"):
            fut.result()


def _print_same_line(msg: str, *, new_line_end: bool = False) -> None:
    msg = msg.strip()
    global MAX_LINE_LENGTH
    MAX_LINE_LENGTH = max(len(msg), MAX_LINE_LENGTH)
    empty_space = max(MAX_LINE_LENGTH - len(msg), 0) * " "
    print(msg + empty_space, end="\r" if not new_line_end else "\n")


def _wait_for_successful_ipyparallel_client_start(
    client: Any,
    n: int,
    timeout: int,
) -> Any:
    from ipyparallel.error import NoEnginesRegistered

    n_engines_old = 0
    for t in range(timeout):
        n_engines = len(client)
        with suppress(NoEnginesRegistered):
            # This can happen, we just need to wait a little longer.
            dview = client[:]
        msg = f"Connected to {n_engines} out of {n} engines after {t} seconds."
        _print_same_line(msg, new_line_end=(n_engines_old != n_engines))
        if n_engines >= n:
            return dview
        n_engines_old = n_engines
        time.sleep(1)
    msg = f"Not all ({n_engines}/{n}) connected after {timeout} seconds."
    raise Exception(msg)  # noqa: TRY002


def connect_to_ipyparallel(
    n: int,
    profile: str,
    timeout: int = 300,
    folder: str | None = None,
    client_kwargs: dict | None = None,
) -> Any:
    """Connect to an `ipcluster` on the cluster headnode.

    Parameters
    ----------
    n
        Number of engines to be started.
    profile
        Profile name of IPython profile.
    timeout
        Time for which we try to connect to get all the engines.
    folder
        Folder that is added to the path of the engines, e.g. ``"~/Work/my_current_project"``.
    client_kwargs
        Keyword arguments passed to `ipyparallel.Client`.

    Returns
    -------
    client
        An IPyparallel client.

    """
    from ipyparallel import Client

    client = Client(profile=profile, **(client_kwargs or {}))
    dview = _wait_for_successful_ipyparallel_client_start(client, n, timeout)
    dview.use_dill()

    if folder is not None:
        console.print(f"Adding {folder} to path.")
        cmd = f"import sys, os; sys.path.append(os.path.expanduser('{folder}'))"
        dview.execute(cmd).result()

    return client


def _get_default_args(func: Callable) -> dict[str, str]:
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def log_exception(log: Any, msg: str, exception: Exception) -> None:
    """Log an exception with a message."""
    try:
        raise exception  # noqa: TRY301
    except Exception:
        log.exception(msg)


def _serialize(msg: Any) -> list:
    return [cloudpickle.dumps(msg)]


def _deserialize(frames: list) -> Any:
    try:
        return cloudpickle.loads(frames[0])
    except pickle.UnpicklingError as e:
        if r"\x03" in str(e):
            # Means that the frame is empty because it only contains an end of text char
            # `\x03  ^C    (End of text)`
            # TODO: Not sure why this happens.
            console.log(
                r"pickle.UnpicklingError in _deserialize: Received an empty frame (\x03).",
            )
            console.print_exception(show_locals=True)
        raise


class LRUCachedCallable:
    """Wraps a function to become cached.

    Parameters
    ----------
    function
    max_size
        Cache size of the LRU cache, by default 128.
    with_cloudpickle
        Use cloudpickle for storing the data in memory.

    """

    def __init__(
        self,
        function: Callable[..., Any],
        *,
        max_size: int = 128,
        with_cloudpickle: bool = False,
    ) -> None:
        """Initialize the cache."""
        self.max_size = max_size
        self.function = function
        self._with_cloudpickle = with_cloudpickle
        self._signature = signature(self.function)
        if max_size == 0:
            return
        manager = Manager()
        self._cache_dict = manager.dict()
        self._cache_queue = manager.list()
        self._cache_lock = manager.Lock()

    def _get_from_cache(self, key: str) -> tuple[bool, Any]:
        """Get a value from the cache by key."""
        if self.max_size == 0:
            value = None
        with self._cache_lock:
            value = self._cache_dict.get(key)
            if value is not None:  # Move key to back of queue
                self._cache_queue.remove(key)
                self._cache_queue.append(key)
        if value is not None:
            found = True
            if value == _NONE_RETURN_STR:
                value = None
            elif self._with_cloudpickle:
                value = cloudpickle.loads(value)
        else:
            found = False
        return found, value

    def _insert_into_cache(self, key: str, value: Any) -> ListProxy[Any]:
        """Insert a key value pair into the cache."""
        if value is None:
            value = _NONE_RETURN_STR
        elif self._with_cloudpickle:
            value = cloudpickle.dumps(value)
        with self._cache_lock:
            cache_size = len(self._cache_queue)
            self._cache_dict[key] = value
            if cache_size < self.max_size:
                self._cache_queue.append(key)
            else:
                key_to_evict = self._cache_queue.pop(0)
                self._cache_dict.pop(key_to_evict)
                self._cache_queue.append(key)
            return self._cache_queue

    @property
    def cache_dict(self) -> dict:
        """Returns a copy of the cache."""
        return dict(self._cache_dict.items())

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Call the wrapped function and cache the result."""
        bound_args = self._signature.bind(*args, **kwargs)
        bound_args.apply_defaults()
        if self.max_size == 0:
            return self.function(*args, **kwargs)
        key = str(bound_args.arguments)
        found, value = self._get_from_cache(key)
        if found:
            return value
        ret = self.function(*args, **kwargs)
        self._insert_into_cache(key, ret)
        return ret


def shared_memory_cache(cache_size: int = 128) -> Callable:
    """Create a cache similar to `functools.lru_cache`.

    This will actually cache the return values of the function, whereas
    `functools.lru_cache` will pickle the decorated function each time
    with an empty cache.
    """

    def cache_decorator(function: Callable[..., Any]) -> Callable[..., Any]:
        return functools.wraps(function)(
            LRUCachedCallable(function, max_size=cache_size),
        )

    return cache_decorator


def _prefix(
    fname: str | list[str] | Path | list[Path],
) -> str:
    if isinstance(fname, tuple | list):
        return f".{len(fname):08}_learners."
    if isinstance(fname, str | Path):
        return ".learner."
    msg = "Incorrect type for fname."
    raise TypeError(msg)


def fname_to_learner_fname(
    fname: str | list[str] | Path | list[Path],
) -> Path:
    """Convert a learner filename (data) to a filename used to cloudpickle the learner."""
    prefix = _prefix(fname)
    if isinstance(fname, tuple | list):
        fname = fname[0]
    p = Path(fname)
    new_name = f"{prefix}{p.stem}{p.suffix}"
    return p.with_name(new_name)


def fname_to_learner(
    fname: str | list[str] | Path | list[Path],
    *,
    return_initializer: bool = False,
) -> tuple[adaptive.BaseLearner, Callable[[], None] | None] | adaptive.BaseLearner:
    """Load a learner from a filename (based on cloudpickled learner)."""
    learner_name = fname_to_learner_fname(fname)
    with learner_name.open("rb") as f:
        learner, initializer = cloudpickle.load(f)
    if return_initializer:
        return learner, initializer
    return learner


def _ensure_folder_exists(
    fnames: FnamesTypes,
) -> None:
    if isinstance(fnames[0], tuple | list):
        for _fnames in fnames:
            assert isinstance(_fnames, tuple | list)
            _ensure_folder_exists(_fnames)
    else:
        assert isinstance(fnames[0], str | Path)
        folders = {Path(fname).parent for fname in fnames}  # type: ignore[arg-type]
        for folder in folders:
            folder.mkdir(parents=True, exist_ok=True)


def cloudpickle_learners(
    learners: list[adaptive.BaseLearner],
    fnames: FnamesTypes,
    *,
    initializers: list[Callable[[], None]] | None = None,
    with_progress_bar: bool = False,
    empty_copies: bool = True,
) -> tuple[int, float]:
    """Save a list of learners to disk using cloudpickle.

    Returns the total size of the saved files in bytes and the total time.
    """
    _ensure_folder_exists(fnames)
    total_filesize = 0
    t_start = time.time()
    if initializers is None:
        initializers = [None] * len(learners)  # type: ignore[list-item]
    for learner, fname, initializer in _progress(
        zip(learners, fnames, initializers, strict=True),
        with_progress_bar,
        desc="Cloudpickling learners",
    ):
        fname_learner = fname_to_learner_fname(fname)
        if empty_copies:
            _at_least_adaptive_version("0.14.1", "empty_copies")
            learner = learner.new()  # noqa: PLW2901
        with fname_learner.open("wb") as f:
            # Make sure that the learner and initializer are ALWAYS pickled together.
            # Otherwise, the learner function and initializer might not use to the
            # same objects even when referring to the same objects in a namespace.
            cloudpickle.dump((learner, initializer), f)
        filesize = fname_learner.stat().st_size
        total_filesize += filesize
    total_time = time.time() - t_start
    return total_filesize, total_time


def fname_to_dataframe(
    fname: str | list[str] | Path | list[Path],
    format: str = "pickle",  # noqa: A002
) -> Path:
    """Convert a learner filename (data) to a filename is used to save the dataframe."""
    if format == "excel":
        format = "xlsx"  # noqa: A001
    if isinstance(fname, tuple | list):
        fname = fname[0]
    p = Path(fname)
    new_name = f"dataframe.{p.stem}.{format}"
    return p.with_name(new_name)


def save_dataframe(
    fname: str | list[str],
    *,
    format: _DATAFRAME_FORMATS = "pickle",  # noqa: A002
    save_kwargs: dict[str, Any] | None = None,
    expand_dicts: bool = True,
    atomically: bool = True,
    **to_dataframe_kwargs: Any,
) -> Callable[[adaptive.BaseLearner], None]:
    """Save the learner's data to disk as pandas.DataFrame."""
    save_kwargs = save_kwargs or {}

    def save(learner: adaptive.BaseLearner) -> None:
        df = learner.to_dataframe(**to_dataframe_kwargs)
        if expand_dicts:
            df = expand_dict_columns(df)
        fname_df = fname_to_dataframe(fname, format=format)

        if format not in get_type_args(_DATAFRAME_FORMATS):
            msg = f"Unknown format {format}"
            raise ValueError(msg)
        do_save = getattr(df, f"to_{format}")

        if format == "hdf":
            assert save_kwargs is not None  # for mypy
            if "key" not in save_kwargs:
                save_kwargs["key"] = "data"

        if atomically:
            with atomic_write(fname_df, return_path=True) as fname_temp:
                do_save(fname_temp, **save_kwargs)
        else:
            do_save(fname_df, **save_kwargs)

    return save


@contextmanager
def atomic_write(
    dest: os.PathLike,
    mode: str = "w",
    *args: Any,
    return_path: bool = False,
    **kwargs: Any,
) -> Any:
    """Write atomically to 'dest', using a temporary file in the same directory.

    This function has the same signature as 'open', except that the default
    mode is 'w', not 'r', and there is an additional keyword-only parameter,
    'return_path'. If 'return_path=True' then a Path pointing to the (as yet
    nonexistant) temporary file is yielded, rather than a file handle.
    This is useful when calling libraries that expect a path, rather than an
    open file handle.
    """
    temp_dest = Path(dest).with_suffix(f".temp.{os.getpid()}.{uuid.uuid4()}")
    try:
        # First create an empty file; this ensures we have the same semantics
        # as 'open(..., mode="w")'.
        temp_dest.open("w").close()
        # Now give control back to the caller.
        if return_path:
            yield temp_dest
        else:
            with temp_dest.open(mode, *args, **kwargs) as fp:
                yield fp
        # Atomically change 'dest' to point to the 'temp_dest' inode.
        os.replace(temp_dest, dest)  # noqa: PTH105
    except Exception:
        with suppress(FileNotFoundError):
            os.remove(temp_dest)  # noqa: PTH107
        raise


_DATAFRAME_FORMATS = Literal[
    "parquet",
    "csv",
    "hdf",
    "pickle",
    "feather",
    "excel",
    "json",
]


def expand_dict_columns(df: pd.DataFrame) -> pd.DataFrame:
    """Expand dict columns in a dataframe."""
    if df.empty:
        return df

    for col, val in df.iloc[0].items():
        if isinstance(val, dict):
            prefix = f"{col}."
            x = pd.json_normalize(df.pop(col)).add_prefix(prefix)  # type: ignore[arg-type]
            x.index = df.index
            for _col in x:
                assert _col not in df, f"{_col=} already exists in df."
            df = df.join(x)
    return df


def load_dataframes(
    fnames: list[str] | list[list[str]] | list[Path] | list[list[Path]],
    *,
    concat: bool = True,
    read_kwargs: dict[str, Any] | None = None,
    format: _DATAFRAME_FORMATS = "pickle",  # noqa: A002
) -> pd.DataFrame | list[pd.DataFrame]:
    """Load a list of dataframes from disk."""
    read_kwargs = read_kwargs or {}
    if format == "hdf" and "key" not in read_kwargs:
        read_kwargs["key"] = "data"

    if format not in get_type_args(_DATAFRAME_FORMATS):
        msg = f"Unknown format {format}."
        raise ValueError(msg)

    do_read = getattr(pd, f"read_{format}")

    dfs = []
    for fn in fnames:
        fn_df = fname_to_dataframe(fn, format=format)
        if not os.path.exists(fn_df):  # noqa: PTH110
            continue
        try:
            df = do_read(fn_df, **read_kwargs)
        except Exception:  # noqa: BLE001
            msg = f"`{fn}`'s DataFrame ({fn_df}) could not be read."
            console.print(msg)
            continue

        df["fname"] = len(df) * [fn]
        dfs.append(df)

    if concat:
        if dfs:
            return pd.concat(dfs, axis=0)
        return pd.DataFrame()
    return dfs


def _at_least_adaptive_version(
    version: str,
    name: str = "",
    *,
    raises: bool = True,
) -> bool:
    import pkg_resources

    required = pkg_resources.parse_version(version)
    v = adaptive.__version__
    v_clean = ".".join(v.split(".")[:3])  # remove the dev0 or other suffix
    current = pkg_resources.parse_version(v_clean)
    if current < required:
        if raises:
            msg = (
                f"`{name}` requires adaptive version "
                f"of at least '{required}', currently using '{current}'.",
            )
            raise RuntimeError(msg)
        return False
    return True


class _TimeGoal:
    def __init__(self, dt: timedelta | datetime) -> None:
        self.dt = dt
        self.start_time: datetime | None = None

    def __call__(self, learner: adaptive.BaseLearner) -> bool:  # noqa: ARG002
        if isinstance(self.dt, timedelta):
            if self.start_time is None:
                self.start_time = datetime.now()  # noqa: DTZ005
            return datetime.now() - self.start_time > self.dt  # noqa: DTZ005
        if isinstance(self.dt, datetime):
            return datetime.now() > self.dt  # noqa: DTZ005
        msg = f"{self.dt=} is not a datetime or timedelta."
        raise TypeError(msg)


def smart_goal(
    goal: GoalTypes,
    learners: list[adaptive.BaseLearner],
) -> Callable[[adaptive.BaseLearner], bool]:
    """Extract a goal from the learners.

    Parameters
    ----------
    goal
        Either a typical callable goal, or integer for number of points goal,
        or float for loss goal, or None to automatically determine, or
        `datetime.timedelta` for a time-based goal.
    learners
        List of learners.

    Returns
    -------
    Callable[[adaptive.BaseLearner], bool]

    """
    if callable(goal):
        return goal
    if isinstance(goal, int):
        return lambda learner: learner.npoints >= goal
    if isinstance(goal, float):
        return lambda learner: learner.loss() <= goal
    if isinstance(goal, timedelta | datetime):
        return _TimeGoal(goal)
    if goal is None:
        learner_types = {type(learner) for learner in learners}
        if len(learner_types) > 1:
            msg = "Multiple learner types found."
            raise TypeError(msg)
        if isinstance(learners[0], adaptive.SequenceLearner):
            return adaptive.SequenceLearner.done
        warnings.warn(
            "Goal is None which means the learners continue forever!",
            stacklevel=2,
        )
        return lambda _: False
    msg = "goal must be `callable | float | None`"
    raise ValueError(msg)


def _serialize_to_b64(x: Any) -> str:
    serialized_x = cloudpickle.dumps(x)
    return base64.b64encode(serialized_x).decode("utf-8")


def _deserialize_from_b64(x: str) -> Any:
    bytes_ = base64.b64decode(x)
    return cloudpickle.loads(bytes_)


_GLOBAL_CACHE = {}


class WrappedFunction:
    """A wrapper to allow `cloudpickle.load`ed functions with `ProcessPoolExecutor`.

    A wrapper around a serialized function that handles deserialization and
    caches the deserialized function in the worker process.

    Parameters
    ----------
    function
        The function to be serialized and wrapped.
    mode
        All of the options avoids sending the function to all workers.
        If "random_id", store the serialized function only in the global cache.
        If "file", save the serialized function to a file and store the path
        to the file in the global cache. Only keep the path in this object.
        If "memory", store the full serialized function in the object.

    Attributes
    ----------
    _cache_key
        The key used to access the deserialized function in the global cache.

    Examples
    --------
    >>> import cloudpickle
    >>> def square(x):
    ...     return x * x
    >>> wrapped_function = WrappedFunction(square)
    >>> wrapped_function(4)
    16

    """

    def __init__(
        self,
        function: Callable[..., Any],
        *,
        mode: Literal["memory", "random_id", "file"] = "random_id",
    ) -> None:
        """Initialize WrappedFunction."""
        serialized_function = cloudpickle.dumps(function)
        self.mode = mode

        if mode == "file":
            with tempfile.NamedTemporaryFile(delete=False) as f:
                f.write(serialized_function)
                self._cache_key = f.name
        elif mode == "memory":
            self._cache_key = serialized_function
        elif mode == "random_id":
            assert platform.system() == "Linux"
            name = function.__name__ if hasattr(function, "__name__") else "function"
            self._cache_key = f"{name}_{os.urandom(16).hex()}"
        else:
            msg = f"mode={mode} is not valid."
            raise ValueError(msg)

        # This setting of the cache only works on Linux where the default start method
        # is 'fork'. On MacOS it is 'spawn', so the cache can be populated in __call__.
        global _GLOBAL_CACHE  # noqa: PLW0602
        _GLOBAL_CACHE[self._cache_key] = function

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Call the wrapped function.

        Retrieves the deserialized function from the global cache and calls it
        with the provided arguments and keyword arguments.

        Parameters
        ----------
        *args
            Positional arguments to pass to the deserialized function.
        **kwargs
            Keyword arguments to pass to the deserialized function.

        Returns
        -------
        Any
            The result of calling the deserialized function with the provided
            arguments and keyword arguments.

        """
        global _GLOBAL_CACHE  # noqa: PLW0602

        if self._cache_key not in _GLOBAL_CACHE:
            if self.mode == "file":
                with open(self._cache_key, "rb") as f:  # noqa: PTH123
                    serialized_function = f.read()
                deserialized_function = cloudpickle.loads(serialized_function)
                _GLOBAL_CACHE[self._cache_key] = deserialized_function
            elif self.mode == "memory":
                _GLOBAL_CACHE[self._cache_key] = cloudpickle.loads(self._cache_key)
            elif self.mode == "random_id":
                msg = (
                    "The function was not found in the global cache. "
                    "This is likely because the default start method on "
                    "your system is 'spawn' instead of 'fork'. "
                    "Try setting `mode='file' or `mode='memory'`."
                )
                raise RuntimeError(msg)

        deserialized_function = _GLOBAL_CACHE[self._cache_key]
        return deserialized_function(*args, **kwargs)


def _now() -> str:
    """Return the current time as a string."""
    return datetime.now(tz=timezone.utc).isoformat()


def _from_datetime(dt: str) -> datetime:
    """Return a string representation of a datetime object."""
    return datetime.fromisoformat(dt)


def _time_between(start: str, end: str) -> float:
    """Return the time between two strings representing datetimes."""
    dt_start = _from_datetime(start)
    dt_end = _from_datetime(end)
    return (dt_end - dt_start).total_seconds()


async def sleep_unless_task_is_done(
    task: asyncio.Task,
    sleep_duration: float,
    trigger_event: asyncio.Event | None = None,
) -> bool:
    """Sleep for an interval, unless the task or a trigger event is done earlier."""
    # Create the sleep task separately
    sleep_task = asyncio.create_task(asyncio.sleep(sleep_duration))
    tasks_to_wait = [sleep_task, task]

    # Add the trigger_event to the wait list if provided
    if trigger_event:
        # Create a task for the event wait
        event_task = asyncio.create_task(trigger_event.wait())
        tasks_to_wait.append(event_task)

    # Await all tasks until any one is done
    done, pending = await asyncio.wait(
        tasks_to_wait,
        return_when=asyncio.FIRST_COMPLETED,
    )

    # Cancel only the sleep task if it's pending
    if sleep_task in pending:
        sleep_task.cancel()

    if trigger_event is not None and event_task.done():
        trigger_event.clear()

    return task.done()


def _update_progress_for_paths(
    paths_dict: dict[str, set[Path | tuple[Path, ...]]],
    progress: Progress,
    total_task: TaskID | None,
    task_ids: dict[str, TaskID],
) -> int:
    """Update progress bars for each set of paths."""
    n_completed = _remove_completed_paths(paths_dict)
    total_completed = sum(n_completed.values())
    for key, n_done in n_completed.items():
        progress.update(task_ids[key], advance=n_done)
    if total_task is not None:
        progress.update(total_task, advance=total_completed)
    return total_completed


def _remove_completed_paths(
    paths_dict: dict[str, set[Path | tuple[Path, ...]]],
) -> dict[str, int]:
    n_completed = {}

    for key, paths in paths_dict.items():
        completed_count = 0
        to_discard = set()
        to_add = set()

        for path_unit in paths:
            # Check if it's a single Path or a tuple of Paths
            paths_to_check = [path_unit] if isinstance(path_unit, Path) else path_unit

            # Check if all paths in the path_unit exist
            if all(p.exists() for p in paths_to_check):
                completed_count += 1
                to_discard.add(path_unit)
            elif isinstance(path_unit, tuple):
                exists = {p for p in path_unit if p.exists()}
                if any(exists):
                    to_discard.add(path_unit)
                    to_add.add(tuple(p for p in path_unit if p not in exists))

        n_completed[key] = completed_count
        paths_dict[key] -= to_discard
        paths_dict[key] |= to_add

    return n_completed


async def _track_file_creation_progress(
    paths_dict: dict[str, set[Path | tuple[Path, ...]]],
    progress: Progress,
    interval: float = 1,
) -> None:
    """Asynchronously track and update the progress of file creation.

    Parameters
    ----------
    paths_dict
        A dictionary with keys representing categories and values being sets of file paths to monitor.
    progress
        The Progress object from the rich library for displaying progress.
    interval
        The time interval (in seconds) at which to update the progress. The interval is dynamically
        adjusted to be at least 50 times the time it takes to update the progress. This ensures that
        updating the progress does not take up a significant amount of time.

    """
    # create total_files and add_total_progress before updating paths_dict
    total_files = sum(len(paths) for paths in paths_dict.values())
    add_total_progress = len(paths_dict) > 1
    n_completed = _remove_completed_paths(paths_dict)  # updates paths_dict in-place
    total_done = sum(n_completed.values())
    task_ids: dict[str, TaskID] = {}

    # Add a total progress bar only if there are multiple entries in the dictionary
    total_task = (
        progress.add_task(
            "[cyan bold underline]Total",
            total=total_files,
            completed=total_done,
        )
        if add_total_progress
        else None
    )
    for key, n_done in n_completed.items():
        n_remaining = len(paths_dict.get(key, []))
        task_ids[key] = progress.add_task(
            f"[green]{key}",
            total=n_remaining + n_done,
            completed=n_done,
        )
    try:
        progress.start()  # Start the progress display
        total_processed = 0
        while True:
            t_start = time.time()
            total_processed += _update_progress_for_paths(
                paths_dict,
                progress,
                total_task,
                task_ids,
            )
            if total_processed >= total_files:
                progress.refresh()  # Final refresh to ensure 100%
                break  # Exit loop if all files are processed
            progress.refresh()
            # Sleep for at least 50 times the update time
            t_update = time.time() - t_start
            await asyncio.sleep(max(interval, 50 * t_update))

    finally:
        progress.stop()  # Stop the progress display, regardless of what happens


def track_file_creation_progress(
    paths_dict: dict[str, set[Path | tuple[Path, ...]]],
    interval: int = 1,
) -> asyncio.Task:
    """Initialize and asynchronously track the progress of file creation.

    WARNING: This function modifies the provided dictionary in-place.

    This function sets up an asynchronous monitoring system that periodically
    checks for the existence of specified files or groups of files. Each item
    in the provided dictionary can be a single file (Path object) or a group
    of files (tuple of Path objects). The progress is updated for each file or
    group of files only when all files in the group exist. This allows tracking
    of complex file creation processes where multiple files together constitute
    a single unit of work.

    The tracking occurs at regular intervals, specified by the user, and updates
    individual and, if applicable, total progress bars to reflect the current
    state of file creation. It is particularly useful in environments where files
    are expected to be created over time and need to be monitored collectively.

    Parameters
    ----------
    paths_dict : dict[str, set[Union[Path, Tuple[Path, ...]]]]
        A dictionary with keys representing categories and values being sets of
        file paths (Path objects) or groups of file paths (tuples of Path objects)
        to monitor.
    interval : int
        The time interval (in seconds) at which the progress is updated.

    Returns
    -------
    asyncio.Task
        The asyncio Task object that is tracking the file creation progress.

    Examples
    --------
    >>> paths_dict = {
        "docs": {Path("docs/environment.yml"), (Path("doc1.md"), Path("doc2.md"))},
        "example2": {Path("/path/to/file3"), Path("/path/to/file4")},
    }
    >>> task = track_file_creation_progress(paths_dict)

    """
    get_console().clear_live()  # avoid LiveError, only 1 live render allowed at a time
    columns = (*Progress.get_default_columns(), TimeElapsedColumn())
    progress = Progress(*columns, auto_refresh=False)
    coro = _track_file_creation_progress(paths_dict, progress, interval)
    ioloop = asyncio.get_event_loop()
    return ioloop.create_task(coro)



from __future__ import annotations

from typing import TYPE_CHECKING

import ipywidgets as ipw
from IPython.display import display

from adaptive_scheduler.widgets import _disable_widgets_output_scrollbar, info

if TYPE_CHECKING:
    from adaptive_scheduler._server_support.run_manager import RunManager


class MultiRunManager:
    """A manager that can contain multiple RunManagers.

    Parameters
    ----------
    run_managers
        Initial list of RunManagers to include.

    Attributes
    ----------
    run_managers
        Dictionary of managed RunManagers, keyed by their names.

    """

    def __init__(self, run_managers: list[RunManager] | None = None) -> None:
        self.run_managers: dict[str, RunManager] = {}
        self._widget: ipw.VBox | None = None
        self._tab_widget: ipw.Tab | None = None
        self._info_widgets: dict[str, ipw.Widget] = {}
        self._update_all_button: ipw.Button | None = None
        if run_managers:
            for rm in run_managers:
                self.add_run_manager(rm)

    def add_run_manager(
        self,
        run_manager: RunManager,
        *,
        start: bool = False,
        wait_for: str | None = None,
    ) -> None:
        """Add a new RunManager to the MultiRunManager.

        Parameters
        ----------
        run_manager
            The RunManager to add.
        start
            Whether to start the RunManager immediately after adding it.
        wait_for
            The name of another RunManager to wait for before starting this one.
            Only applicable if start is True.

        Raises
        ------
        ValueError
            If a RunManager with the same name already exists.
        KeyError
            If the specified wait_for RunManager does not exist.

        """
        if run_manager.job_name in self.run_managers:
            msg = f"A RunManager with the name '{run_manager.job_name}' already exists."
            raise ValueError(msg)

        self.run_managers[run_manager.job_name] = run_manager
        self._info_widgets[run_manager.job_name] = info(
            run_manager,
            display_widget=False,
            disable_widgets_output_scrollbar=False,
        )

        if start:
            if wait_for:
                if wait_for not in self.run_managers:
                    msg = f"No RunManager with the name '{wait_for}' exists."
                    raise KeyError(msg)
                run_manager.start(wait_for=self.run_managers[wait_for])
            else:
                run_manager.start()
        elif wait_for:
            msg = "`start` must be True if `wait_for` is used."
            raise ValueError(msg)

        if self._widget is not None:
            self._update_widget()

    def remove_run_manager(self, name: str) -> None:
        """Remove a RunManager from the MultiRunManager.

        Parameters
        ----------
        name
            The name of the RunManager to remove.

        Raises
        ------
        KeyError
            If no RunManager with the given name exists.

        """
        if name in self.run_managers:
            rm = self.run_managers.pop(name)
            rm.cancel()
            self._info_widgets.pop(name)
            if self._widget is not None:
                self._update_widget()
        else:
            msg = f"No RunManager with the name '{name}' exists."
            raise KeyError(msg)

    def start_all(self) -> None:
        """Start all RunManagers."""
        for run_manager in self.run_managers.values():
            run_manager.start()

    def cancel_all(self) -> None:
        """Cancel all RunManagers."""
        for run_manager in self.run_managers.values():
            run_manager.cancel()

    def _create_widget(self) -> ipw.VBox:
        """Create the widget for displaying RunManager info and update button."""
        self._tab_widget = ipw.Tab()
        children = list(self._info_widgets.values())
        self._tab_widget.children = children
        for i, name in enumerate(self.run_managers.keys()):
            self._tab_widget.set_title(i, name)

        self._update_all_button = ipw.Button(
            description="Update All",
            button_style="info",
            tooltip="Update all RunManagers",
        )
        self._update_all_button.on_click(self._update_all_callback)

        return ipw.VBox([self._update_all_button, self._tab_widget])

    def _update_widget(self) -> None:
        """Update the widget when RunManagers are added or removed."""
        if self._tab_widget is not None:
            current_children = list(self._tab_widget.children)
            new_children = list(self._info_widgets.values())

            # Create a new tuple of children
            updated_children = tuple(child for child in current_children if child in new_children)

            # Add new children
            for widget in new_children:
                if widget not in updated_children:
                    updated_children += (widget,)

            # Update the widget's children
            self._tab_widget.children = updated_children

            # Update titles
            for i, name in enumerate(self.run_managers.keys()):
                self._tab_widget.set_title(i, name)

    def _update_all_callback(self, _: ipw.Button) -> None:
        """Callback function for the Update All button."""
        assert self._tab_widget is not None
        for widget in self._tab_widget.children:
            update_button = widget.children[0].children[1].children[0]
            assert update_button.description == "update info"
            update_button.click()

    def info(self) -> ipw.VBox:
        """Display info about all RunManagers in a widget with an Update All button."""
        if self._widget is None:
            _disable_widgets_output_scrollbar()
            self._widget = self._create_widget()
        return self._widget

    def _repr_html_(self) -> str:
        """HTML representation for Jupyter notebooks."""
        return self.info()

    def display(self) -> None:
        """Display the widget."""
        display(self.info())



"""PBS for Adaptive Scheduler."""

from __future__ import annotations

import collections
import getpass
import math
import os
import os.path
import subprocess
import textwrap
import warnings
from pathlib import Path
from typing import TYPE_CHECKING

from adaptive_scheduler._scheduler.base_scheduler import BaseScheduler
from adaptive_scheduler._scheduler.common import run_submit

if TYPE_CHECKING:
    from typing import Any

    from adaptive_scheduler.utils import EXECUTOR_TYPES


class PBS(BaseScheduler):
    """PBS scheduler."""

    # Attributes that all schedulers need to have
    _ext = ".batch"
    # the "-k oe" flags with "qsub" writes the log output to
    # files directly instead of at the end of the job. The downside
    # is that the logfiles are put in the homefolder.
    _submit_cmd = "qsub -k oe"
    _JOB_ID_VARIABLE = "${PBS_JOBID}"
    _options_flag = "PBS"
    _cancel_cmd = "qdel"

    def __init__(
        self,
        cores: int,
        *,
        python_executable: str | None = None,
        log_folder: str | Path = "",
        mpiexec_executable: str | None = None,
        executor_type: EXECUTOR_TYPES = "process-pool",
        num_threads: int = 1,
        extra_scheduler: list[str] | None = None,
        extra_env_vars: list[str] | None = None,
        extra_script: str | None = None,
        batch_folder: str | Path = "",
        # Extra PBS specific arguments
        cores_per_node: int | None = None,
    ) -> None:
        """Initialize the PBS scheduler."""
        super().__init__(
            cores,
            python_executable=python_executable,
            log_folder=log_folder,
            mpiexec_executable=mpiexec_executable,
            executor_type=executor_type,
            num_threads=num_threads,
            extra_scheduler=extra_scheduler,
            extra_env_vars=extra_env_vars,
            extra_script=extra_script,
            batch_folder=batch_folder,
        )
        # PBS specific
        self.cores_per_node = cores_per_node
        self._calculate_nnodes()
        if cores != self.cores:
            warnings.warn(
                f"`self.cores` changed from {cores} to {self.cores}",
                stacklevel=2,
            )

    def __getstate__(self) -> dict[str, Any]:
        """Return the state of the scheduler."""
        # PBS has one different argument from the BaseScheduler
        return dict(**super().__getstate__(), cores_per_node=self.cores_per_node)

    @staticmethod
    def sanatize_job_id(job_id: str) -> str:
        """Changes '91722.hpc05.hpc' into '91722'."""
        return job_id.split(".")[0]

    def _calculate_nnodes(self) -> None:
        assert isinstance(self.cores, int), "self.cores must be an integer for PBS."
        if self.cores_per_node is None:
            partial_msg = "Use set `cores_per_node=...` before passing the scheduler."
            try:
                max_cores_per_node = self._guess_cores_per_node()
                self.nnodes = math.ceil(self.cores / max_cores_per_node)
                self.cores_per_node = round(self.cores / self.nnodes)
                msg = (
                    f"`#PBS -l nodes={self.nnodes}:ppn={self.cores_per_node}` is"
                    f" guessed using the `qnodes` command, we set"
                    f" `cores_per_node={self.cores_per_node}`."
                    f" You might want to change this. {partial_msg}"
                )
                warnings.warn(msg, stacklevel=2)
                self.cores = self.nnodes * self.cores_per_node
            except Exception as e:  # noqa: BLE001
                msg = (
                    f"Got an error: {e}."
                    " Couldn't guess `cores_per_node`, this argument is required"
                    f" for PBS. {partial_msg}"
                    " We set `cores_per_node=1`!"
                )
                warnings.warn(msg, stacklevel=2)
                self.nnodes = self.cores
                self.cores_per_nodes = 1
        else:
            self.nnodes = self.cores / self.cores_per_node  # type: ignore[assignment]
            if not float(self.nnodes).is_integer():
                msg = "cores / cores_per_node must be an integer!"
                raise ValueError(msg)
            self.nnodes = int(self.nnodes)

    def output_fnames(self, name: str) -> list[Path]:
        """Get the output filenames."""
        # The "-k oe" flags with "qsub" writes the log output to
        # files directly instead of at the end of the job. The downside
        # is that the logfiles are put in the home folder.
        home = Path.home()
        stdout, stderr = (home / f"{name}.{x}{self._JOB_ID_VARIABLE}" for x in "oe")
        return [stdout, stderr]

    def job_script(self, options: dict[str, Any], *, index: int | None = None) -> str:
        """Get a jobscript in string form.

        Returns
        -------
        job_script
            A job script that can be submitted to PBS.
        index
            The index of the job that is being run. This is used when
            specifying different resources for different jobs.
            Currently not implemented for PBS!

        """
        job_script = textwrap.dedent(
            f"""\
            #!/bin/sh
            #PBS -l nodes={self.nnodes}:ppn={self.cores_per_node}
            #PBS -V
            #PBS -o /tmp/placeholder
            #PBS -e /tmp/placeholder
            {{extra_scheduler}}

            {{extra_env_vars}}

            cd $PBS_O_WORKDIR

            {{extra_script}}

            {{executor_specific}}
            """,
        )

        return job_script.format(
            extra_scheduler=self.extra_scheduler(index=index),
            extra_env_vars=self.extra_env_vars(index=index),
            extra_script=self.extra_script(index=index),
            executor_specific=self._executor_specific("${NAME}", options, index=index),
            job_id_variable=self._JOB_ID_VARIABLE,
        )

    def start_job(self, name: str, *, index: int | None = None) -> None:
        """Writes a job script and submits it to the scheduler."""
        if index is not None:
            msg = "PBS does not support `index`."
            raise NotImplementedError(msg)
        name_prefix = name.rsplit("-", 1)[0]
        name_opt = f"-N {name}"
        submit_cmd = f"{self.submit_cmd} {name_opt} {self.batch_fname(name_prefix)}"
        run_submit(submit_cmd, name)

    @staticmethod
    def _split_by_job(lines: list[str]) -> list[list[str]]:
        jobs: list[list[str]] = [[]]
        for line in lines:
            line = line.strip()  # noqa: PLW2901
            if line:
                jobs[-1].append(line)
            else:
                jobs.append([])
        return [j for j in jobs if j]

    @staticmethod
    def _fix_line_cuts(raw_info: list[str]) -> list[str]:
        info = []
        for line in raw_info:
            if " = " in line:
                info.append(line)
            else:
                info[-1] += line
        return info

    def queue(self, *, me_only: bool = True) -> dict[str, dict]:
        """Get the status of all jobs in the queue."""
        cmd = ["qstat", "-f"]

        proc = subprocess.run(
            cmd,
            text=True,
            capture_output=True,
            check=False,
            env=dict(os.environ, SGE_LONG_QNAMES="1000"),
        )
        output = proc.stdout

        if proc.returncode != 0:
            msg = "qstat is not responding."
            raise RuntimeError(msg)

        jobs = self._split_by_job(output.replace("\n\t", "").split("\n"))

        running = {}
        for header, *raw_info in jobs:
            job_id = header.split("Job Id: ")[1]
            info = dict([line.split(" = ") for line in self._fix_line_cuts(raw_info)])
            if info["job_state"] in ["R", "Q"]:
                info["job_name"] = info["Job_Name"]  # used in `server_support.manage_jobs`
                info["state"] = info["job_state"]  # used in `RunManager.live`
                running[job_id] = info

        if me_only:
            # We do this because the "-u [username here]"  flag doesn't
            # work with "-f" on some clusters.
            username = getpass.getuser()
            running = {
                job_id: info for job_id, info in running.items() if username in info["Job_Owner"]
            }

        return running

    def _qnodes(self) -> dict[str, dict[str, str]]:
        proc = subprocess.run(["qnodes"], text=True, capture_output=True, check=False)
        output = proc.stdout

        if proc.returncode != 0:
            msg = "qnodes is not responding."
            raise RuntimeError(msg)

        jobs = self._split_by_job(output.replace("\n\t", "").split("\n"))

        return {
            node: dict([line.split(" = ") for line in self._fix_line_cuts(raw_info)])
            for node, *raw_info in jobs
        }

    def _guess_cores_per_node(self) -> int:
        nodes = self._qnodes()
        cntr = collections.Counter([int(info["np"]) for info in nodes.values()])
        ncores, freq = cntr.most_common(1)[0]
        return ncores



from __future__ import annotations

import os

if os.environ.get("EXECUTOR_TYPE") == "dask-mpi":
    from dask_mpi import initialize

    initialize()

import argparse
import os
from concurrent.futures import ProcessPoolExecutor
from contextlib import suppress
from typing import TYPE_CHECKING, Any, get_args

import adaptive
import cloudpickle

from adaptive_scheduler import client_support
from adaptive_scheduler.utils import (
    _DATAFRAME_FORMATS,
    LOKY_START_METHODS,
    WrappedFunction,
    _deserialize_from_b64,
)

if TYPE_CHECKING:
    from adaptive_scheduler.utils import EXECUTOR_TYPES


if os.environ.get("EXECUTOR_TYPE") == "mpi4py":
    from mpi4py import MPI

    MPI.pickle.__init__(cloudpickle.dumps, cloudpickle.loads)


class SequentialExecutor(ProcessPoolExecutor):
    """A asynchronous executor that runs tasks sequentially."""

    def __init__(self) -> None:
        super().__init__(max_workers=1)


def _get_executor(
    executor_type: EXECUTOR_TYPES,
    profile: str | None,
    n: int,
    loky_start_method: LOKY_START_METHODS | None,
) -> Any:
    if executor_type == "mpi4py":
        from mpi4py.futures import MPIPoolExecutor

        return MPIPoolExecutor()
    if executor_type == "ipyparallel":
        from adaptive_scheduler.utils import connect_to_ipyparallel

        assert profile is not None
        return connect_to_ipyparallel(profile=profile, n=n)
    if executor_type == "dask-mpi":
        from distributed import Client

        return Client()
    if executor_type == "loky":
        import loky

        assert loky_start_method is not None
        loky.backend.context.set_start_method(loky_start_method)
        return loky.get_reusable_executor(max_workers=n)
    if executor_type == "process-pool":
        from concurrent.futures import ProcessPoolExecutor

        return ProcessPoolExecutor(max_workers=n)
    if executor_type == "sequential":
        return SequentialExecutor()
    msg = f"Unknown executor_type: {executor_type}"
    raise ValueError(msg)


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("--profile", action="store", type=str, default=None)
    parser.add_argument("--n", action="store", dest="n", type=int)
    parser.add_argument("--log-fname", action="store", type=str)
    parser.add_argument("--job-id", action="store", type=str)
    parser.add_argument("--name", action="store", dest="name", type=str, required=True)
    parser.add_argument("--url", action="store", type=str, required=True)

    parser.add_argument("--save-dataframe", action="store_true", default=False)
    parser.add_argument(
        "--dataframe-format",
        action="store",
        type=str,
        choices=get_args(_DATAFRAME_FORMATS),
    )
    parser.add_argument(
        "--executor-type",
        action="store",
        type=str,
        default="process-pool",
    )
    parser.add_argument(
        "--loky-start-method",
        action="store",
        type=str,
        default="loky",
        choices=get_args(LOKY_START_METHODS),
    )
    parser.add_argument("--log-interval", action="store", type=float, default=60)
    parser.add_argument(
        "--save-interval",
        action="store",
        type=float,
        default=120,
    )
    parser.add_argument("--serialized-runner-kwargs", action="store", type=str)
    return parser.parse_args()


def main() -> None:
    """The main function that is called by the launcher script."""
    args = _parse_args()
    client_support.add_log_file_handler(args.log_fname)
    client_support.log.info("parsed args", **vars(args))
    client_support.args_to_env(args)

    # ask the server for a fname and learner
    learner, fname, initializer = client_support.get_learner(
        args.url,
        args.log_fname,
        args.job_id,
        args.name,
    )

    if initializer is not None:
        initializer()

    if args.executor_type in ("process-pool", "sequential"):
        learner.function = WrappedFunction(learner.function)

    with suppress(Exception):
        learner.load(fname)
    npoints_start = learner.npoints

    executor = _get_executor(
        args.executor_type,
        args.profile,
        args.n,
        args.loky_start_method,
    )

    runner_kwargs = _deserialize_from_b64(args.serialized_runner_kwargs)

    runner_kwargs.setdefault("shutdown_executor", True)
    runner = adaptive.Runner(learner, executor=executor, **runner_kwargs)

    # periodically save the data (in case the job dies)
    runner.start_periodic_saving({"fname": fname}, interval=args.save_interval)

    if args.save_dataframe:
        from adaptive_scheduler.utils import save_dataframe

        save_method = save_dataframe(fname, format=args.dataframe_format)
        runner.start_periodic_saving(interval=args.save_interval, method=save_method)

    # log progress info in the job output script, optional
    _log_task = client_support.log_info(runner, interval=args.log_interval)

    # block until runner goal reached
    runner.ioloop.run_until_complete(runner.task)

    # save once more after the runner is done
    learner.save(fname)

    if args.save_dataframe:
        save_method(learner)

    # log once more after the runner is done
    client_support.log_now(runner, npoints_start)

    # tell the database that this learner has reached its goal
    client_support.tell_done(args.url, fname)


if __name__ == "__main__":
    main()



"""Tests for the SlurmExecutor class."""

from __future__ import annotations

import asyncio
import time
from typing import TYPE_CHECKING, Any
from unittest.mock import patch

import adaptive
import pytest

from adaptive_scheduler import SlurmExecutor
from adaptive_scheduler._server_support.run_manager import RunManager

if TYPE_CHECKING:
    from pathlib import Path


def example_func(x: float) -> float:
    """Example function that returns its input."""
    return x


@pytest.fixture
def executor(tmp_path: Path) -> SlurmExecutor:
    """Create a SlurmExecutor instance."""
    return SlurmExecutor(
        name="test",
        folder=tmp_path,
        save_interval=1,
        log_interval=1,
        job_manager_interval=1,
    )


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_submit_single_task(executor: SlurmExecutor) -> None:
    """Test submitting a single task."""
    task = executor.submit(example_func, 1.0, 2.0)
    assert task.task_id.learner_index == 0
    assert task.task_id.sequence_index == 0
    assert executor._sequence_mapping[example_func] == 0
    assert executor._sequences[example_func] == [(1.0, 2.0)]


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_submit_multiple_tasks_same_function(executor: SlurmExecutor) -> None:
    """Test submitting multiple tasks with the same function."""
    tasks = [executor.submit(example_func, x) for x in [1.0, 2.0, 3.0]]
    assert all(task.task_id.learner_index == 0 for task in tasks)
    assert [task.task_id.sequence_index for task in tasks] == [0, 1, 2]
    assert executor._sequence_mapping[example_func] == 0
    assert executor._sequences[example_func] == [(1.0,), (2.0,), (3.0,)]


def another_func(x: float) -> float:
    """Another example function."""
    return x * 2


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_submit_multiple_functions(executor: SlurmExecutor) -> None:
    """Test submitting tasks with different functions."""
    task1 = executor.submit(example_func, 1.0)
    task2 = executor.submit(another_func, 2.0)

    assert task1.task_id.learner_index == 0
    assert task2.task_id.learner_index == 1
    assert len(executor._sequence_mapping) == 2
    assert len(executor._sequences) == 2


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_submit_with_kwargs_raises(executor: SlurmExecutor) -> None:
    """Test that submitting with kwargs raises ValueError."""
    with pytest.raises(ValueError, match="Keyword arguments are not supported"):
        executor.submit(example_func, x=1.0)


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_finalize_creates_run_manager(executor: SlurmExecutor) -> None:
    """Test that finalize creates a RunManager."""
    executor.submit(example_func, 1.0)
    rm = executor.finalize(start=False)
    assert isinstance(rm, RunManager)
    assert len(rm.learners) == 1
    assert len(rm.fnames) == 1
    assert isinstance(rm, RunManager)


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_finalize_twice_raises(executor: SlurmExecutor) -> None:
    """Test that calling finalize twice raises RuntimeError."""
    executor.submit(example_func, 1.0)
    executor.finalize(start=False)
    with pytest.raises(RuntimeError, match="RunManager already initialized"):
        executor.finalize()


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_map(executor: SlurmExecutor) -> None:
    """Test the map method."""
    values = [1.0, 2.0, 3.0]
    tasks = executor.map(example_func, values)
    assert len(tasks) == len(values)
    assert all(task.task_id.learner_index == 0 for task in tasks)  # type: ignore[attr-defined]
    assert [task.task_id.sequence_index for task in tasks] == [0, 1, 2]  # type: ignore[attr-defined]


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_map_with_timeout_raises(executor: SlurmExecutor) -> None:
    """Test that map with timeout raises NotImplementedError."""
    with pytest.raises(NotImplementedError, match="Timeout not implemented"):
        executor.map(example_func, [1.0], timeout=1.0)


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_map_with_chunksize_raises(executor: SlurmExecutor) -> None:
    """Test that map with chunksize raises NotImplementedError."""
    with pytest.raises(NotImplementedError, match="Chunksize not implemented"):
        executor.map(example_func, [1.0], chunksize=2)


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_new_executor(executor: SlurmExecutor) -> None:
    """Test creating a new executor with the same parameters."""
    executor.submit(example_func, 1.0)
    executor.finalize(start=False)

    new_executor = executor.new()
    assert new_executor._run_manager is None
    assert not new_executor._sequences
    assert not new_executor._sequence_mapping
    assert new_executor.name == executor.name
    assert new_executor.folder == executor.folder


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_folder_creation(tmp_path: Path) -> None:
    """Test that the folder is created correctly."""
    folder = tmp_path / "test_folder"
    executor = SlurmExecutor(folder=folder)
    assert executor.folder == folder

    # Test default folder creation
    executor = SlurmExecutor()
    assert executor.folder.parent.name == ".adaptive_scheduler"  # type: ignore[union-attr]
    assert len(executor.folder.name.split("-")) == 3  # type: ignore[union-attr]
    # date-time-uuid format


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_cleanup(executor: SlurmExecutor) -> None:
    """Test the cleanup method."""
    executor.submit(example_func, 1.0)
    executor.finalize(start=False)

    executor.cleanup()


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_task_get_before_finalize(executor: SlurmExecutor) -> None:
    """Test that _get before finalize returns None."""
    task = executor.submit(example_func, 1.0)
    with pytest.raises(RuntimeError, match="Task mapping not found; finalize()"):
        task._get()


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_task_get_with_missing_file(executor: SlurmExecutor) -> None:
    """Test that _get with missing file returns None."""
    task = executor.submit(example_func, 1.0)
    executor.finalize(start=False)
    assert task._get() is None  # File doesn't exist yet


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_task_get(executor: SlurmExecutor) -> None:
    """Test that _get gets the data."""
    task = executor.submit(example_func, 1.0)
    executor.finalize(start=False)

    # First _get should try to load
    assert task._get() is None

    # Create a dummy file
    learner, fname = task._learner_and_fname
    adaptive.runner.simple(learner)
    learner.save(fname)

    # Does not respect min_load_interval because data is already in memory
    assert task._get() == 1.0


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_task_get_adapts_load_interval(executor: SlurmExecutor) -> None:
    """Test that _get adapts min_load_interval based on load time."""
    task = executor.submit(example_func, 1.0)
    executor.finalize(start=False)

    # Mock a slow load operation
    def slow_load(*args: Any) -> None:  # noqa: ARG001
        time.sleep(0.1)  # Simulate slow load

    learner, fname = task._learner_and_fname
    with patch.object(learner, "load", side_effect=slow_load):
        # Create a dummy file
        with open(fname, "wb") as f:  # noqa: PTH123
            f.write(b"dummy")

        task._get()
        assert task.min_load_interval >= 2.0  # Should be at least 20x the load time


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_task_result_timeout_not_implemented(executor: SlurmExecutor) -> None:
    """Test that result() with timeout raises NotImplementedError."""
    task = executor.submit(example_func, 1.0)
    with pytest.raises(NotImplementedError, match="Timeout not implemented"):
        task.result(timeout=1.0)


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_task_repr_triggers_get(executor: SlurmExecutor) -> None:
    """Test that repr triggers _get."""
    task = executor.submit(example_func, 1.0)
    executor.finalize(start=False)

    assert "PENDING" in repr(task)

    # Mock learner.done() to return True
    learner, _ = task._learner_and_fname
    with patch.object(learner, "done", return_value=True), patch.dict(learner.data, {0: 42}):
        assert "FINISHED" in repr(task)


@pytest.mark.asyncio
@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
async def test_task_await(executor: SlurmExecutor) -> None:
    """Test awaiting a task."""
    task = executor.submit(example_func, 1.0)
    executor.finalize(start=False)

    # Create a background task to simulate result appearing
    async def simulate_result() -> None:
        await asyncio.sleep(0.1)
        learner, fname = task._learner_and_fname
        learner.data[0] = 42
        with patch.object(learner, "done", return_value=True):
            task._get()

    asyncio.create_task(simulate_result())  # noqa: RUF006
    result = await task
    assert result == 42


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_to_learners_mapping_single_function(tmp_path: Path) -> None:
    """Test that _to_learners creates the correct mapping for a single function."""
    executor = SlurmExecutor(folder=tmp_path, size_per_learner=2)
    # Submit 5 tasks to example_func so that they are split into chunks of 2.
    for i in range(5):
        executor.submit(example_func, i)
    learners, fnames, mapping = executor._to_learners()

    # We expect ceil(5/2) = 3 learners.
    assert len(learners) == 3

    func_id = executor._sequence_mapping[example_func]
    expected_mapping = {
        (func_id, 0): (0, 0),  # first learner, first task
        (func_id, 1): (0, 1),  # first learner, second task
        (func_id, 2): (1, 0),  # second learner, first task
        (func_id, 3): (1, 1),  # second learner, second task
        (func_id, 4): (2, 0),  # third learner, first task (only one task in this chunk)
    }
    assert mapping == expected_mapping


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_finalize_mapping_and_learners(tmp_path: Path) -> None:
    """Test that finalize() sets the task mapping correctly and creates the right number of learners."""
    executor = SlurmExecutor(folder=tmp_path, size_per_learner=2)
    # Submit 3 tasks to example_func.
    for i in range(3):
        executor.submit(example_func, i)

    rm = executor.finalize(start=False)
    # For 3 tasks with chunk size 2:
    #   - The first chunk (learner 0) has tasks 0 and 1.
    #   - The second chunk (learner 1) has task 2.
    func_id = executor._sequence_mapping[example_func]
    expected_mapping = {
        (func_id, 0): (0, 0),
        (func_id, 1): (0, 1),
        (func_id, 2): (1, 0),
    }
    assert executor._task_mapping == expected_mapping
    # Also, the run manager should have 2 learners.
    assert isinstance(rm, RunManager)
    assert len(rm.learners) == 2


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_task_get_with_chunking(tmp_path: Path) -> None:
    """Test that tasks in different learners retrieve the correct result when using size_per_learner."""
    executor = SlurmExecutor(folder=tmp_path, size_per_learner=2, save_interval=1)
    # Submit three tasks; with size_per_learner=2, this will produce 2 learners.
    task1 = executor.submit(example_func, 42)
    task2 = executor.submit(example_func, 43)
    task3 = executor.submit(example_func, 44)
    rm = executor.finalize(start=False)

    # For learner 0 (tasks 0 and 1)
    assert isinstance(rm, RunManager)
    learner0 = rm.learners[0]
    fname0 = rm.fnames[0]
    learner0.data[0] = 42
    learner0.data[1] = 43
    learner0.save(fname0)
    # For learner 1 (task 2)
    learner1 = rm.learners[1]
    fname1 = rm.fnames[1]
    learner1.data[0] = 44
    learner1.save(fname1)

    # _get() should now retrieve the correct values based on the mapping.
    assert task1._get() == 42
    assert task2._get() == 43
    assert task3._get() == 44


@pytest.mark.usefixtures("_mock_slurm_partitions")
@pytest.mark.usefixtures("_mock_slurm_queue")
def test_mapping_multiple_functions(tmp_path: Path) -> None:
    """Test that the mapping is correct when tasks are submitted for multiple functions."""
    executor = SlurmExecutor(folder=tmp_path, size_per_learner=2)
    # Submit two tasks for example_func and two for another_func.
    executor.submit(example_func, 10)
    executor.submit(example_func, 20)
    executor.submit(another_func, 5)
    executor.submit(another_func, 6)

    # Directly call _to_learners to examine the mapping.
    learners, fnames, mapping = executor._to_learners()

    expected_mapping = {
        # For example_func: two tasks in one learner (since 2 tasks fit in one chunk).
        (executor._sequence_mapping[example_func], 0): (0, 0),
        (executor._sequence_mapping[example_func], 1): (0, 1),
        # For another_func: two tasks in one learner.
        (executor._sequence_mapping[another_func], 0): (1, 0),
        (executor._sequence_mapping[another_func], 1): (1, 1),
    }
    assert mapping == expected_mapping
    assert len(learners) == 2



"""Test the RunManager class."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, cast
from unittest.mock import patch

import adaptive
import pytest
from ipywidgets import VBox

from adaptive_scheduler._server_support.run_manager import (
    RunManager,
    start_one_by_one,
)

from .helpers import (
    MockScheduler,
    get_socket,
    send_message,
    temporary_working_directory,
)

if TYPE_CHECKING:
    from pathlib import Path


def test_run_manager_init(
    mock_scheduler: MockScheduler,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test the initialization of RunManager."""
    rm = RunManager(mock_scheduler, learners, fnames)
    assert isinstance(rm, RunManager)


@pytest.mark.asyncio
async def test_run_manager_start_and_cancel(
    mock_scheduler: MockScheduler,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test starting the RunManager."""
    rm = RunManager(mock_scheduler, learners, fnames)
    rm.start()
    assert rm.status() == "running"
    rm.cancel()
    await asyncio.sleep(0.1)
    assert rm.status() == "cancelled"


def test_run_manager_cleanup(
    mock_scheduler: MockScheduler,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
    tmp_path: Path,
) -> None:
    """Test the cleanup method of RunManager."""
    with temporary_working_directory(tmp_path):
        rm = RunManager(mock_scheduler, learners, fnames)
        assert rm.move_old_logs_to is not None
        rm.move_old_logs_to.mkdir(parents=True, exist_ok=True)
        rm.cleanup(remove_old_logs_folder=True)
        assert not rm.move_old_logs_to.exists()


def test_run_manager_parse_log_files(
    mock_scheduler: MockScheduler,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test parsing log files in RunManager."""
    rm = RunManager(mock_scheduler, learners, fnames)
    rm.start()
    df = rm.parse_log_files(only_last=True)
    assert df.empty  # nothing has been run yet


def test_run_manager_load_learners(
    mock_scheduler: MockScheduler,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test loading learners in RunManager."""
    rm = RunManager(mock_scheduler, learners, fnames)
    for lrn, fn in zip(learners, fnames, strict=True):
        adaptive.runner.simple(lrn, npoints_goal=10)
        lrn.save(fn)
    rm.load_learners()
    for learner in learners:
        assert learner.data


@pytest.mark.asyncio
async def test_run_manager_elapsed_time(
    mock_scheduler: MockScheduler,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test elapsed time in RunManager."""
    rm = RunManager(mock_scheduler, learners, fnames)
    assert rm.elapsed_time() == 0
    rm.start()
    await asyncio.sleep(0.05)
    assert rm.elapsed_time() > 0
    rm.cancel()
    await asyncio.sleep(0.05)
    assert rm.elapsed_time() > 0


@pytest.mark.asyncio
async def test_run_manager_status(
    mock_scheduler: MockScheduler,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test the status of RunManager."""
    rm = RunManager(mock_scheduler, learners, fnames)
    assert rm.status() == "not yet started"
    rm.start()
    await asyncio.sleep(0.05)
    assert rm.status() == "running"
    rm.cancel()
    await asyncio.sleep(0.05)
    assert rm.status() == "cancelled"


def test_run_manager_repr_html(
    mock_scheduler: MockScheduler,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test the _repr_html_ method of RunManager."""
    rm = RunManager(mock_scheduler, learners, fnames)
    rm.start()
    with patch("IPython.display.display") as mocked_display:
        rm._repr_html_()
        assert mocked_display.called


def test_run_manager_info(
    mock_scheduler: MockScheduler,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test the info method of RunManager."""
    rm = RunManager(mock_scheduler, learners, fnames)
    rm.start()
    with patch("IPython.display.display") as mocked_display:
        rm.info()
        assert mocked_display.called
        display_arg = mocked_display.call_args[0]
        # Check the content of the display
        assert isinstance(display_arg, tuple)
        assert isinstance(display_arg[0], VBox)


def test_run_manager_load_dataframes(
    mock_scheduler: MockScheduler,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test loading dataframes in RunManager."""
    rm = RunManager(mock_scheduler, learners, fnames, save_dataframe=False)
    with pytest.raises(
        ValueError,
        match="The `save_dataframe` option was not set to True.",
    ):
        rm.load_dataframes()


@pytest.mark.asyncio
async def test_start_one_by_one(
    mock_scheduler: MockScheduler,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test starting RunManagers one by one."""

    def goal(rm: RunManager) -> bool:
        return rm.elapsed_time() >= 2

    rm1 = RunManager(mock_scheduler, learners[:1], fnames[:1], job_name="rm1")
    rm2 = RunManager(mock_scheduler, learners[1:], fnames[1:], job_name="rm2")

    rm1.start()
    tasks = start_one_by_one(rm1, rm2, goal=goal)
    await asyncio.sleep(0.2)
    assert isinstance(tasks, tuple)
    assert len(tasks) == 2
    assert isinstance(tasks[0], asyncio.Future)
    assert isinstance(tasks[1], list)
    assert isinstance(tasks[1][0], asyncio.Future)

    # Cancel tasks
    tasks[0].cancel()
    tasks[1][0].cancel()
    rm1.cancel()
    rm2.cancel()


@pytest.mark.asyncio
async def test_run_manager_auto_restart(
    mock_scheduler: MockScheduler,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test starting the RunManager."""
    rm = RunManager(mock_scheduler, learners, fnames, job_manager_interval=0.1)
    rm.start()
    await asyncio.sleep(0.1)
    assert rm.task is not None
    assert rm.status() == "running"
    q = rm.scheduler.queue()
    assert "0" in q
    assert "1" in q
    # Note: we don't know whether it is:
    # This: "0": {"job_name": "adaptive-scheduler-0"}, "1": {"job_name": "adaptive-scheduler-1"}
    # or:   "1": {"job_name": "adaptive-scheduler-0"}, "0": {"job_name": "adaptive-scheduler-1"}
    job_names = job_name0, job_name1 = "adaptive-scheduler-0", "adaptive-scheduler-1"
    job_id0, job_id1 = rm.scheduler.job_names_to_job_ids(job_name0, job_name1)
    log_fnames = ("log0.log", "log1.log")

    db = rm.database_manager.as_dicts()
    # jobs were all assigned...
    assert all(x["job_name"] is not None for x in db)
    # but they have not yet reported that they started.
    assert all(x["job_id"] is None for x in db)

    # Send a start message to the DatabaseManager
    # This is coming from the client
    with get_socket(rm.database_manager) as socket:
        for i, (entry, log_fname) in enumerate(zip(db, log_fnames, strict=True)):
            job_id = str(i)
            job_name = entry["job_name"]
            start_message = ("start", job_id, log_fname, job_name)
            await send_message(socket, start_message)

    db = rm.database_manager.as_dicts()
    assert len(db) == 2
    for entry in db:
        assert entry["job_id"] in ("0", "1")
        assert entry["job_name"] in job_names
        assert not entry["is_done"]

    # Mark the first job as cancelled
    rm.scheduler.update_queue(job_name0, "C")  # type: ignore[attr-defined]
    assert rm.scheduler.queue() == {
        job_id0: {"job_name": job_name0, "status": "C"},
        job_id1: {
            "job_name": "adaptive-scheduler-1",
            "status": "R",
            "state": "RUNNING",
        },
    }
    cast("MockScheduler", rm.scheduler)._queue_info.pop(job_id0)
    await asyncio.sleep(0.15)
    # Check that the job is restarted automatically with a new job_id:
    q = rm.scheduler.queue()
    assert q["2"] == {"job_name": job_name0, "state": "RUNNING", "status": "R"}

    # Start the job "from the client"
    with get_socket(rm.database_manager) as socket:
        start_message = ("start", "2", "log2.log", job_name0)
        await send_message(socket, start_message)

    await asyncio.sleep(0.1)

    # Check if the new job is started in the database
    db = rm.database_manager.as_dicts()
    assert len(db) == 2
    for entry in db:
        assert entry["job_id"] in ("0", "1", "2")
        assert entry["job_name"] in job_names
        assert not entry["is_done"]
        assert entry["log_fname"] == f"log{entry['job_id']}.log"

    # Now mark the 2 jobs as done
    with get_socket(rm.database_manager) as socket:
        for entry in db:
            stop_message = ("stop", entry["fname"])
            await send_message(socket, stop_message)

    await asyncio.sleep(0.15)

    # Check that the jobs are now done
    db = rm.database_manager.as_dicts()
    assert len(db) == 2
    for entry in db:
        assert entry["job_id"] is None
        assert entry["job_name"] is None
        assert entry["is_done"]
        assert entry["log_fname"].endswith(".log")

    # Check that RunManager is done
    assert rm.task.done()



from __future__ import annotations

import asyncio
import shutil
import time
import warnings
import weakref
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pandas as pd

from adaptive_scheduler.utils import (
    LOKY_START_METHODS,
    GoalTypes,
    _at_least_adaptive_version,
    _remove_or_move_files,
    _time_between,
    fname_to_learner_fname,
    load_dataframes,
    load_parallel,
    sleep_unless_task_is_done,
    smart_goal,
)
from adaptive_scheduler.widgets import info

from .base_manager import BaseManager
from .common import (
    _delete_old_ipython_profiles,
    _maybe_path,
    cleanup_scheduler_files,
    console,
    get_allowed_url,
)
from .database_manager import DatabaseManager
from .job_manager import JobManager
from .kill_manager import KillManager
from .parse_logs import parse_log_files

if TYPE_CHECKING:
    from collections.abc import Callable

    import adaptive

    from adaptive_scheduler.scheduler import BaseScheduler
    from adaptive_scheduler.utils import _DATAFRAME_FORMATS


# This is a global list of all active run managers.
ACTIVE_RUN_MANAGERS: list[weakref.ReferenceType[RunManager]] = []


class RunManager(BaseManager):
    """A convenience tool that starts the job, database, and kill manager.

    Parameters
    ----------
    scheduler
        A scheduler instance from `adaptive_scheduler.scheduler`.
    learners
        List of `learners` corresponding to `fnames`.
    fnames
        List of `fnames` corresponding to `learners`.
    goal
        The goal passed to the `adaptive.Runner`. Note that this function will
        be serialized and pasted in the ``job_script``. Can be a smart-goal
        that accepts
        ``Callable[[adaptive.BaseLearner], bool] | float | datetime | timedelta | None``.
        See `adaptive_scheduler.utils.smart_goal` for more information.
    check_goal_on_start
        Checks whether a learner is already done. Only works if the learner is loaded.
    dependencies
        Dictionary of dependencies, e.g., ``{1: [0]}`` means that the ``learners[1]``
        depends on the ``learners[0]``. This means that the ``learners[1]`` will only
        start when the ``learners[0]`` is done.
    runner_kwargs
        Extra keyword argument to pass to the `adaptive.Runner`. Note that this dict
        will be serialized and pasted in the ``job_script``.
    url
        The url of the database manager, with the format
        ``tcp://ip_of_this_machine:allowed_port.``. If None, a correct url will be chosen.
    save_interval
        Time in seconds between saving of the learners.
    log_interval
        Time in seconds between log entries.
    job_name
        From this string the job names will be created, e.g.
        ``["adaptive-scheduler-1", "adaptive-scheduler-2", ...]``.
    job_manager_interval
        Time in seconds between checking and starting jobs.
    kill_interval
        Check for `kill_on_error` string inside the log-files every `kill_interval` seconds.
    kill_on_error
        If ``error`` is a string and is found in the log files, the job will
        be cancelled and restarted. If it is a callable, it is applied
        to the log text. Must take a single argument, a list of
        strings, and return True if the job has to be killed, or
        False if not. Set to None if no `KillManager` is needed.
    move_old_logs_to
        Move logs of killed jobs to this directory. If None the logs will be deleted.
    db_fname
        Filename of the database, e.g. 'running.json'.
    overwrite_db
        Overwrite the existing database.
    job_manager_kwargs
        Keyword arguments for the `JobManager` function that aren't set in ``__init__`` here.
    kill_manager_kwargs
        Keyword arguments for the `KillManager` function that aren't set in ``__init__`` here.
    loky_start_method
        Loky start method, by default "loky".
    cleanup_first
        Cancel all previous jobs generated by the same RunManager and clean logfiles.
    save_dataframe
        Whether to periodically save the learner's data as a `pandas.DataFame`.
    dataframe_format
        The format in which to save the `pandas.DataFame`. See the type hint for the options.
    max_log_lines
        The maximum number of lines to display in the log viewer widget.
    max_fails_per_job
        Maximum number of times that a job can fail. This is here as a fail switch
        because a job might fail instantly because of a bug inside your code.
        The job manager will stop when
        ``n_jobs * total_number_of_jobs_failed > max_fails_per_job`` is true.
    max_simultaneous_jobs
        Maximum number of simultaneously running jobs. By default no more than 500
        jobs will be running. Keep in mind that if you do not specify a ``runner.goal``,
        jobs will run forever, resulting in the jobs that were not initially started
        (because of this `max_simultaneous_jobs` condition) to not ever start.
    initializers
        List of functions that are called before the job starts, can populate
        a cache.
    quiet
        Whether to show a progress bar when creating learner files.

    Attributes
    ----------
    job_names : list
        List of job_names. Generated with ``self.job_name``.
    database_manager : `DatabaseManager`
        The database manager.
    job_manager : `JobManager`
        The job manager.
    kill_manager : `KillManager` or None
        The kill manager.
    start_time : float or None
        Time at which ``self.start()`` is called.
    end_time : float or None
        Time at which the jobs are all done or at which ``self.cancel()`` is called.

    Examples
    --------
    Here is an example of using the `RunManager` with a modified ``job_script_function``.

    >>> import adaptive_scheduler
    >>> scheduler = adaptive_scheduler.scheduler.SLURM(cores=10)
    >>> run_manager = adaptive_scheduler.server_support.RunManager(
    ...     scheduler=scheduler
    ... ).start()

    Or an example using `ipyparallel.Client`.

    >>> from functools import partial
    >>> import adaptive_scheduler
    >>> scheduler = adaptive_scheduler.scheduler.SLURM(
    ...     cores=10, executor_type="ipyparallel",
    ... )
    >>> def goal(learner):
    ...     return learner.npoints > 2000
    >>> run_manager = adaptive_scheduler.server_support.RunManager(
    ...     scheduler=scheduler,
    ...     goal=goal,
    ...     log_interval=30,
    ...     save_interval=30,
    ... )
    >>> run_manager.start()

    """

    def __init__(  # noqa: PLR0915
        self,
        scheduler: BaseScheduler,
        learners: list[adaptive.BaseLearner],
        fnames: list[str] | list[Path],
        *,
        goal: GoalTypes | None = None,
        check_goal_on_start: bool = True,
        dependencies: dict[int, list[int]] | None = None,
        runner_kwargs: dict | None = None,
        url: str | None = None,
        save_interval: float = 300,
        log_interval: float = 300,
        job_name: str = "adaptive-scheduler",
        job_manager_interval: float = 60,
        kill_interval: float = 60,
        kill_on_error: str | Callable[[list[str]], bool] | None = "srun: error:",
        move_old_logs_to: str | Path | None = "old_logs",
        db_fname: str | Path | None = None,
        overwrite_db: bool = True,
        job_manager_kwargs: dict[str, Any] | None = None,
        kill_manager_kwargs: dict[str, Any] | None = None,
        loky_start_method: LOKY_START_METHODS = "loky",
        cleanup_first: bool = False,
        save_dataframe: bool = False,
        dataframe_format: _DATAFRAME_FORMATS = "pickle",
        max_log_lines: int = 500,
        max_fails_per_job: int = 50,
        max_simultaneous_jobs: int = 100,
        initializers: list[Callable[[], None]] | None = None,
        quiet: bool = False,
    ) -> None:
        super().__init__()

        # Set from arguments
        self.scheduler = scheduler
        self.goal = smart_goal(goal, learners)
        self.dependencies = dependencies
        self.check_goal_on_start = check_goal_on_start
        self.runner_kwargs = runner_kwargs
        self.save_interval = save_interval
        self.log_interval = log_interval
        self.job_name = job_name
        self.job_manager_interval = job_manager_interval
        self.kill_interval = kill_interval
        self.kill_on_error = kill_on_error
        self.move_old_logs_to = _maybe_path(move_old_logs_to)
        self.db_fname = Path(db_fname or f"{job_name}-database.json")
        self.overwrite_db = overwrite_db
        self.job_manager_kwargs = job_manager_kwargs or {}
        self.kill_manager_kwargs = kill_manager_kwargs or {}
        self.loky_start_method = loky_start_method
        self.save_dataframe = save_dataframe
        self.dataframe_format = dataframe_format
        self.max_log_lines = max_log_lines
        self.max_fails_per_job = max_fails_per_job
        self.max_simultaneous_jobs = max_simultaneous_jobs
        self.initializers = initializers
        self.quiet = quiet
        # Track job start times, (job_name, start_time) -> request_time
        self._job_start_time_dict: dict[tuple[str, str], str] = {}

        for key in ["max_fails_per_job", "max_simultaneous_jobs"]:
            if key in self.job_manager_kwargs:
                msg = (
                    f"The `{key}` argument is not allowed in `job_manager_kwargs`."
                    " Please specify it in `RunManager.__init__` instead.",
                )
                raise ValueError(msg)

        if self.save_dataframe:
            _at_least_adaptive_version("0.14.0", "save_dataframe")

        # Set in methods
        self.start_time: float | None = None
        self.end_time: float | None = None
        self._start_one_by_one_task: tuple[asyncio.Future, list[asyncio.Task]] | None = None

        # Set on init
        self.learners = learners
        self.fnames = fnames

        self._last_load_time: dict[int, float] = {}

        if isinstance(self.fnames[0], list | tuple):
            # For a BalancingLearner
            assert isinstance(self.fnames[0][0], str | Path)
        else:
            assert isinstance(self.fnames[0], str | Path)

        self.job_names = [f"{self.job_name}-{i}" for i in range(len(self.learners))]

        if cleanup_first:
            self.scheduler.cancel(self.job_names)
            self.cleanup(remove_old_logs_folder=True)

        self.url = url or get_allowed_url()
        self.database_manager = DatabaseManager(
            url=self.url,
            scheduler=self.scheduler,
            db_fname=self.db_fname,
            learners=self.learners,
            fnames=self.fnames,
            dependencies=self.dependencies,
            overwrite_db=self.overwrite_db,
            initializers=self.initializers,
            with_progress_bar=not self.quiet,
        )
        self.job_manager = JobManager(
            self.job_names,
            self.database_manager,
            scheduler=self.scheduler,
            interval=self.job_manager_interval,
            max_fails_per_job=self.max_fails_per_job,
            max_simultaneous_jobs=self.max_simultaneous_jobs,
            # Launcher command line options
            save_dataframe=self.save_dataframe,
            dataframe_format=self.dataframe_format,
            loky_start_method=self.loky_start_method,
            log_interval=self.log_interval,
            save_interval=self.save_interval,
            runner_kwargs=self.runner_kwargs,
            goal=self.goal,
            **self.job_manager_kwargs,
        )
        self.kill_manager: KillManager | None
        if self.kill_on_error is not None:
            self.kill_manager = KillManager(
                scheduler=self.scheduler,
                database_manager=self.database_manager,
                error=self.kill_on_error,
                interval=self.kill_interval,
                move_to=self.move_old_logs_to,
                **self.kill_manager_kwargs,
            )
        else:
            self.kill_manager = None

    def _setup(self) -> None:
        self.database_manager.start()
        if self.check_goal_on_start:
            # Check if goal already reached
            # Only works after the `database_manager` has started.
            done_fnames = [
                fname
                for fname, learner in zip(self.fnames, self.learners, strict=True)
                if self.goal(learner)
            ]
            self.database_manager._stop_requests(done_fnames)  # type: ignore[arg-type]
        self.job_manager.start()
        if self.kill_manager:
            self.kill_manager.start()
        self.start_time = time.time()

    def start(self, wait_for: RunManager | None = None) -> RunManager:  # type: ignore[override]
        """Start the RunManager and optionally wait for another RunManager to finish."""
        _track_and_maybe_cancel_existing(self)
        if wait_for is not None:
            self._start_one_by_one_task = start_one_by_one(wait_for, self)
        else:
            super().start()
        return self

    async def _manage(self) -> None:
        assert self.job_manager.task is not None
        while not self.job_manager.task.done():
            if self.job_manager._request_times:
                for job in self.database_manager.as_dicts():
                    start_time = job["start_time"]
                    job_name = job["job_name"]
                    # Check if the job actually started (not cancelled)
                    if (
                        start_time is not None
                        and job_name in self.job_manager._request_times
                        and (job_name, start_time) not in self._job_start_time_dict
                    ):
                        request_time = self.job_manager._request_times.pop(job_name)
                        self._job_start_time_dict[job_name, start_time] = request_time

            if await sleep_unless_task_is_done(
                self.database_manager.task,  # type: ignore[arg-type]
                5,
            ):  # if true, we are done
                break

        self.end_time = time.time()

    def job_starting_times(self) -> list[float]:
        """Return the starting times of the jobs."""
        return [_time_between(end, start) for (_, start), end in self._job_start_time_dict.items()]

    def cancel(self) -> None:
        """Cancel the manager tasks and the jobs in the queue."""
        self.database_manager.cancel()
        self.job_manager.cancel()
        if self.kill_manager is not None:
            self.kill_manager.cancel()
        self.scheduler.cancel(self.job_names)
        if self.task is not None:
            self.task.cancel()
        self.end_time = time.time()
        if self._start_one_by_one_task is not None:
            self._start_one_by_one_task[0].cancel()

    def cleanup(self, *, remove_old_logs_folder: bool = False) -> None:
        """Cleanup the log and batch files."""
        for fname in self.fnames:
            fname_cloudpickle = fname_to_learner_fname(fname)
            with suppress(FileNotFoundError):
                fname_cloudpickle.unlink()
        if "ipyparallel" in self.scheduler.executor_type:
            # executor_type is a tuple or string
            _delete_old_ipython_profiles(self.scheduler)

        cleanup_scheduler_files(
            job_names=self.job_names,
            scheduler=self.scheduler,
            with_progress_bar=not self.quiet,
            move_to=self.move_old_logs_to,
        )
        if remove_old_logs_folder and self.move_old_logs_to is not None:
            with suppress(FileNotFoundError):
                shutil.rmtree(self.move_old_logs_to)

    def parse_log_files(self, *, only_last: bool = True) -> pd.DataFrame:
        """Parse the log-files and convert it to a `~pandas.core.frame.DataFrame`.

        Parameters
        ----------
        only_last
            Only look use the last printed status message.

        Returns
        -------
        pandas.core.frame.DataFrame

        """
        return parse_log_files(
            self.database_manager,
            self.scheduler,
            only_last=only_last,
        )

    def task_status(self) -> None:
        r"""Print the stack of the `asyncio.Task`\s."""
        if self.job_manager.task is not None:
            self.job_manager.task.print_stack()
        if self.database_manager.task is not None:
            self.database_manager.task.print_stack()
        if self.kill_manager is not None and self.kill_manager.task is not None:
            self.kill_manager.task.print_stack()
        if self.task is not None:
            self.task.print_stack()

    def get_database(self) -> pd.DataFrame:
        """Get the database as a `pandas.DataFrame`."""
        return pd.DataFrame(self.database_manager.as_dicts())

    def load_learners(self) -> None:
        """Load the learners in parallel using `adaptive_scheduler.utils.load_parallel`."""
        t = time.monotonic()
        for i in range(len(self.learners)):
            self._last_load_time[i] = t
        load_parallel(self.learners, self.fnames)

    def elapsed_time(self) -> float:
        """Total time elapsed since the `RunManager` was started."""
        if not self.is_started:
            return 0
        assert self.job_manager.task is not None  # for mypy
        if self.job_manager.task.done():
            end_time = self.end_time
            if end_time is None:
                # task was cancelled before it began
                assert self.job_manager.task.cancelled()
                return 0
        else:
            end_time = time.time()
        return end_time - self.start_time  # type: ignore[operator]

    def status(self) -> str:
        """Return the current status of the `RunManager`."""
        if not self.is_started:
            return "not yet started"

        try:
            assert self.job_manager.task is not None
            self.job_manager.task.result()
        except asyncio.InvalidStateError:
            status = "running"
        except asyncio.CancelledError:
            status = "cancelled"
        except Exception:  # noqa: BLE001
            status = "failed"
            console.log("`JobManager` failed because of the following")
            console.print_exception(show_locals=True)
        else:
            status = "finished"

        try:
            assert self.database_manager.task is not None  # for mypy
            self.database_manager.task.result()
        except (asyncio.InvalidStateError, asyncio.CancelledError):
            pass
        except Exception:  # noqa: BLE001
            status = "failed"
            console.log("`DatabaseManager` failed because of the following")
            console.print_exception(show_locals=True)
        if status == "running":
            return "running"

        if self.end_time is None:
            self.end_time = time.time()
        return status

    def _repr_html_(self) -> None:
        return self.info()

    def info(self) -> None:
        return info(self, display_widget=True)

    def load_dataframes(self) -> pd.DataFrame:
        """Load the `pandas.DataFrame`s with the most recently saved learners data."""
        if not self.save_dataframe:
            msg = "The `save_dataframe` option was not set to True."
            raise ValueError(msg)
        return load_dataframes(self.fnames, format=self.dataframe_format)  # type: ignore[return-value]

    def remove_existing_data(
        self,
        *,
        move_to: str | Path | None = None,
        force: bool = False,
    ) -> None:
        """Remove the existing data files.

        Parameters
        ----------
        move_to
            Move the files to this directory. If None the files will be deleted.
        force
            Remove the files even if the `RunManager` is already started.

        """
        if self.status() == "running" and not force:
            msg = (
                "Data files can only be removed if the `RunManager` is not running."
                " Use `force=True` to remove them anyway."
            )
            raise RuntimeError(msg)

        _remove_or_move_files(self.fnames, move_to=move_to)


async def _wait_for_finished(
    manager_first: RunManager,
    manager_second: RunManager,
    goal: Callable[[RunManager], bool] | None = None,
    interval: float = 120,
) -> None:
    if goal is None:
        assert manager_first.task is not None  # for mpypy
        await manager_first.task
    else:
        while not goal(manager_first):  # noqa: ASYNC110
            await asyncio.sleep(interval)
    manager_second.start()


def _start_after(
    manager_first: RunManager,
    manager_second: RunManager,
    goal: Callable[[RunManager], bool] | None = None,
    interval: float = 120,
) -> asyncio.Task:
    if manager_second.is_started:
        msg = "The second manager must not be started yet."
        raise ValueError(msg)
    coro = _wait_for_finished(manager_first, manager_second, goal, interval)
    return asyncio.create_task(coro)


def start_one_by_one(
    *run_managers: RunManager,
    goal: Callable[[RunManager], bool] | None = None,
    interval: float = 120,
) -> tuple[asyncio.Future, list[asyncio.Task]]:
    """Start a list of RunManagers after each other.

    Parameters
    ----------
    run_managers
        A list of RunManagers.
    goal
        A callable that takes a RunManager as argument and returns a boolean.
        If `goal` is not None, the RunManagers will be started after `goal`
        returns True for the previous RunManager. If `goal` is None, the
        RunManagers will be started after the previous RunManager has finished.
    interval
        The interval at which to check if `goal` is True. Only used if `goal`
        is not None.

    Returns
    -------
    tuple[asyncio.Future, list[asyncio.Future]]
        The first element is the grouped task that starts all RunManagers.
        The second element is a list of tasks that start each RunManager.

    Examples
    --------
    >>> manager_1 = adaptive_scheduler.slurm_run(
    ...     learners[:5],
    ...     fnames[:5],
    ...     partition="hb120rsv2-low",
    ...     goal=0.01,
    ...     name="first",
    ... )
    >>> manager_1.start()
    >>> manager_2 = adaptive_scheduler.slurm_run(
    ...     learners[5:],
    ...     fnames[5:],
    ...     partition="hb120rsv2-low",
    ...     goal=0.01,
    ...     name="second",
    ... )
    >>> # Start second when the first RunManager has more than 1000 points.
    >>> def start_goal(run_manager):
    ...     df = run_manager.parse_log_files()
    ...     npoints = df.get("npoints")
    ...     if npoints is None:
    ...         return False
    ...     return npoints.sum() > 1000
    >>> tasks = adaptive_scheduler.start_one_by_one(
    ...     manager_1,
    ...     manager_2,
    ...     goal=start_goal,
    ... )

    """
    uniques = ["job_name", "db_fname"]
    for u in uniques:
        if len({getattr(r, u) for r in run_managers}) != len(run_managers):
            msg = (
                f"All `RunManager`s must have a unique {u}."
                " If using `slurm_run` these are controlled through the `name` argument.",
            )
            raise ValueError(msg)

    tasks = [
        _start_after(run_managers[i], run_managers[i + 1], goal, interval)
        for i in range(len(run_managers) - 1)
    ]
    return asyncio.gather(*tasks), tasks


def _track_and_maybe_cancel_existing(current_run_manager: RunManager) -> None:
    """Cancel the existing run manager with the same name.

    Additionally, update the ACTIVE_RUN_MANAGERS list with the new run manager.
    Emit a warning when canceling another run manager.
    """
    # First, cancel and remove the old run manager(s) with the same job_name
    to_cancel = [
        weak_rm
        for weak_rm in ACTIVE_RUN_MANAGERS
        if weak_rm() is not None
        and weak_rm().job_name == current_run_manager.job_name  # type: ignore[union-attr]
        and weak_rm() is not current_run_manager  # type: ignore[union-attr]
    ]

    for weak_rm in to_cancel:
        rm = weak_rm()
        if rm is not None:
            # Emit a warning before canceling
            warnings.warn(
                f"Cancelling a RunManager with job_name '{rm.job_name}' because"
                " another instance is created with the same name.",
                UserWarning,
                stacklevel=2,
            )
            rm.cancel()
            ACTIVE_RUN_MANAGERS.remove(weak_rm)

    # Then, check if the current run manager is already in the list
    current_weak_rm = weakref.ref(current_run_manager)
    if not any(current_weak_rm() is rm() for rm in ACTIVE_RUN_MANAGERS):
        ACTIVE_RUN_MANAGERS.append(current_weak_rm)

    # Cleanup: Remove all run_managers that are done, cancelled, or have been garbage collected
    ACTIVE_RUN_MANAGERS[:] = [
        rm
        for rm in ACTIVE_RUN_MANAGERS
        if rm() is not None and rm().end_time is None  # type: ignore[union-attr]
    ]
